# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT
import datasets

from nlubridge.nlu_dataset import NluDataset


def from_huggingface(
    hugging_ds,
    has_intents=True,
    has_entities=False,
    toks_col="tokens",
    iob_col="ner_tags",
) -> NluDataset:
    """
    Convert a Huggingface dataset to an NLUdataset.

    Requirements if has_entities is True:
        * hugging_ds uses IOB entity annotation format, with tokens
          specified under tok_col (default: 'tokens') and IOB tags
          specified under column iob_col (default: 'ner_tags')
        * hugging_ds provides a ClassLabel object for the IOB tags
          accessible through hugging_ds.features[iob_col].feature

    Requirements if has_intents is True:
        * hugging_ds stores intent information under column 'label'
        * hugging_ds provides a ClassLabel object for the intents
          accessible through hugging_ds.features['label']
        * if has_entities is False, the text to classify needs to be
          stored under column 'text'

    Compatible datasets (examples):
        * Intent: "banking77", "nlu_evaluation_data",
          "snips_built_in_intents"
        * Entities: "wnut_17", "germaner"

    :param hugging_ds: Huggingface dataset from which to import the
        data
    :type hugging_ds: datasets.arrow_dataset.Dataset
    :param has_intents: If True, will try to retrieve an intent label
        for each case
    :type has_intents: bool
    :param has_entities: If True, will try to retrieve entities from an
        attribute that keeps this information in IOB format
    :type has_entities: bool
    :param toks_col: hugging_ds attribute under which the tokenized
        document is found
    :type toks_col: str
    :param iob_col: hugging_ds attribute under which the IOB labels are
        found
    :type iob_col: str
    """
    if isinstance(hugging_ds, datasets.dataset_dict.DatasetDict):
        raise TypeError("Invalid type for argument hugging_ds. Try picking a split.")
    if not isinstance(hugging_ds, datasets.arrow_dataset.Dataset):
        raise TypeError("Invalid type for argument hugging_ds.")

    if has_intents:
        # get intents for each example
        try:
            label_mapper = hugging_ds.features["label"]
        except KeyError:
            raise ValueError(
                "hugging_ds must provide intent classes in a ClassLabel feature named "
                "'label'"
            )
        intents = [label_mapper.int2str(x) for x in hugging_ds["label"]]
    else:
        intents = None  # type: ignore[assignment]

    if has_entities:
        # get entities and texts generated from token lists for each example
        try:
            token_seqs = hugging_ds[toks_col]
        except KeyError:
            raise ValueError(
                "Expect entities in IOB format, which requires a column with tokenized "
                "text. If this is not stored under column 'tokens', provide the column "
                "name as tok_col."
            )
        try:
            ner_tag_seqs = hugging_ds[iob_col]
        except KeyError:
            raise ValueError(
                "Expect entities in IOB format, which requires a column with IOB tags."
                "If this is not stored under column 'ner_tags', provide the column "
                "name as iob_col."
            )
        try:
            tag_mapper = hugging_ds.features[iob_col].feature
        except KeyError:
            raise ValueError(
                "hugging_ds must provide entity tags in a ClassLabel accessible "
                "through the '{}' feature".format(iob_col)
            )
        texts, entities = iob2dict_batch(token_seqs, ner_tag_seqs, tag_mapper)
    else:
        entities = None

    if not has_entities:
        # get text for each example if not yet generated from tokens
        try:
            texts = hugging_ds["text"]
        except KeyError:
            raise ValueError("hugging_ds must have a 'text' feature")

    ds = NluDataset(texts, intents, entities)

    return ds


def iob2dict_batch(token_seqs, iob_seqs, tag_mapper):
    """Run iob2dict for many docs."""
    all_texts = []
    all_entities = []
    for tokens, ner_tags in zip(token_seqs, iob_seqs):
        text, entities = iob2dict(tokens, ner_tags, tag_mapper)
        all_texts.append(text)
        all_entities.append(entities)
    return all_texts, all_entities


def iob2dict(tokens, ner_tags, tag_mapper):
    """
    Convert an IOB labeled doc to dictionary format.

    :param tokens: List of tokens in the document
    :type tokens: List[str]
    :param ner_tags: List of IOB tags
    :type ner_tags: List[str]
    :param tag_mapper: object holding the IOB tag details. Has method
        int2str() which converts the tag to a readable string like
        `B-Person`.
    """

    def need_blank(token, len_text):
        if (
            token.startswith("'")
            or (token in [":", ".", ",", ";", "?", "!", "..."])
            or (len_text == 0)
        ):
            return False
        else:
            return True

    entities = []
    text = ""
    is_in = False

    for token, tag in zip(tokens, ner_tags):
        # remove trailing blank if certain characters follow
        if need_blank(token, len(text)):
            text += " "

        # init or update entity based on tag prefix
        tag = tag_mapper.int2str(tag)
        if tag.startswith("B-"):
            entity = {"start": len(text), "entity": tag.replace("B-", "")}
            is_in = True
        elif tag.startswith("I-"):
            pass  # nothing to do
        elif (tag == "O") and is_in:
            entity["end"] = len(text) - need_blank(token, len(text))
            entities.append(entity)
            is_in = False
        elif tag == "O":
            pass  # nothing to do

        # append current token to text
        text += token

    # make sure entities at the end of an utterance are completed
    if not tag == "O":
        entity["end"] = len(text)
        entities.append(entity)

    return text, entities



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

from ..nlu_dataset import NluDataset


class Vendor:
    """Abstract class for a vendor."""

    @property
    def name(self):
        """Return the class's name as a hint to the vendor name."""
        return self.__class__.__name__

    @property
    def alias(self):
        """Return alias, which is a customizable name of the instance."""
        return self._alias

    @alias.setter
    def alias(self, value):
        """Set alias, which is a customizable name of the instance."""
        self._alias = value

    def train(self, dataset: NluDataset):
        """
        Train intent and/or entity classification.

        :param dataset: Training data
        """
        raise NotImplementedError

    def test(self, dataset: NluDataset) -> NluDataset:
        """
        Test a given dataset and obtain classification results as NLUdataset.

        The returned NLUdataset will include intent and/or entity predictions, depending
        on what the model can handle and has been trained on.

        :param dataset: Input dataset to be tested
        :return: NLUdataset object comprising the classification results
        """
        raise NotImplementedError

    def train_intent(self, dataset):
        """Train intent classification."""
        raise NotImplementedError

    def test_intent(self, dataset):
        """Test intent classification."""
        raise NotImplementedError



import json
import os

from test_datasets import FIXTURE_PATH


def test_from_huggingface_intents():
    from datasets import ClassLabel, Dataset

    from nlubridge import from_huggingface

    with open(
        os.path.join(FIXTURE_PATH, "huggingface_banking77.json"), "r", encoding="utf-8"
    ) as f:
        testdata = json.load(f)

    hf_ds = Dataset.from_dict(testdata["data"])
    hf_ds.features["label"] = ClassLabel(
        num_classes=len(testdata["label_class_names"]),
        names=testdata["label_class_names"],
    )
    ds = from_huggingface(hf_ds, has_intents=True, has_entities=False)

    assert len(ds) == 8
    assert len(ds.intents) == len(ds)
    assert all(len(x) > 0 for x in ds.intents)
    assert len(ds.texts) == len(ds)
    assert all(len(x) > 0 for x in ds.texts)
    assert len(ds.entities) == len(ds)
    assert all(len(x) == 0 for x in ds.entities)
    assert ds.texts[0] == "I am still waiting on my card?"
    assert ds.texts[3] == "What is my money worth in other countries?"
    assert ds.intents[0] == "card_arrival"
    assert ds.intents[3] == "exchange_rate"


def test_from_huggingface_entities():
    from datasets import ClassLabel, Dataset

    from nlubridge import from_huggingface

    with open(
        os.path.join(FIXTURE_PATH, "huggingface_wnut_17.json"), "r", encoding="utf-8"
    ) as f:
        testdata = json.load(f)

    hf_ds = Dataset.from_dict(testdata["data"])
    hf_ds.features["ner_tags"].feature = ClassLabel(
        num_classes=len(testdata["ner_tags_class_names"]),
        names=testdata["ner_tags_class_names"],
    )

    ds = from_huggingface(hf_ds, has_intents=False, has_entities=True)

    assert len(ds) == 8
    assert len(ds.intents) == len(ds)
    assert all(x is None for x in ds.intents)
    assert len(ds.texts) == len(ds)
    assert all(len(x) > 0 for x in ds.texts)
    assert len(ds.entities) == len(ds)
    assert all(len(x) >= 0 for x in ds.entities)
    assert ds.texts[3] == "today is my last day at the office."
    assert len(ds.entities[0]) == 2
    assert ds.entities[3] == []
    assert ds.entities[4] == [{"start": 0, "end": 7, "entity": "person"}]
    assert ds.entities[6] == [{"start": 18, "end": 43, "entity": "product"}]


def test_from_watson():
    from nlubridge import from_watson_assistant

    ds = from_watson_assistant(os.path.join(FIXTURE_PATH, "watson_intents_export.csv"))
    assert len(ds) == 4
    assert ds.texts[0][0] != '"'
    assert ds.texts[2] == "testing umläuts"
    assert ds.intents[0] == "intent1"
    assert ds.intents[1] == "intent2"


def test_from_rasa_json():
    from nlubridge import from_rasa

    ds = from_rasa(os.path.join(FIXTURE_PATH, "rasa_nlu.json"), format="json")
    assert len(ds) == 5
    assert ds.texts[1] == "testing umläuts"
    assert ds.intents[0] == "restaurant_search"
    assert ds.intents[3] == "affirm"
    assert ds.entities[0] == []
    assert ds.entities[1] == [
        {"start": 31, "end": 36, "value": "north", "entity": "location"}
    ]
    assert len(ds.entities[4]) == 2
    # make sure the custom "role" key is in dataset
    assert ds.entities[4][0].get("role", False)
    # check assumptions about indexes hold
    idx1 = ds.entities[4][0]["start"]
    idx2 = ds.entities[4][0]["end"]
    value = ds.entities[4][0]["value"]
    assert ds.texts[4][idx1:idx2] == value


def test_from_rasa3_yml():
    from nlubridge import from_rasa

    ds = from_rasa(os.path.join(FIXTURE_PATH, "rasa3_nlu.yml"))
    assert len(ds) == 5
    assert ds.texts[1] == "testing umläuts"
    assert ds.intents[0] == "restaurant_search"
    assert ds.intents[4] == "affirm"
    assert ds.entities[0] == []
    assert ds.entities[1] == []
    assert len(ds.entities[3]) == 2
    # make sure the custom "role" key is in dataset
    assert ds.entities[3][1].get("role", False)
    # check assumptions about indexes hold
    idx1 = ds.entities[3][1]["start"]
    idx2 = ds.entities[3][1]["end"]
    value = ds.entities[3][1]["value"]
    assert ds.texts[3][idx1:idx2] == value


def test_from_luis():
    from nlubridge import from_luis

    ds = from_luis(os.path.join(FIXTURE_PATH, "luis_nlu.json"))
    assert len(ds) == 5
    assert ds.texts[1] == "testing umläuts"
    assert ds.intents[0] == "restaurant_search"
    assert ds.intents[3] == "affirm"
    assert ds.entities[0] == []
    assert ds.entities[1] == [
        {"start": 31, "end": 36, "value": "north", "entity": "location"}
    ]
    assert len(ds.entities[4]) == 2
    # make sure the custom "role" key is in dataset
    assert ds.entities[4][0].get("role", False)
    # check assumptions about indexes hold
    idx1 = ds.entities[4][0]["start"]
    idx2 = ds.entities[4][0]["end"]
    value = ds.entities[4][0]["value"]
    assert ds.texts[4][idx1:idx2] == value



# Copyright (c) 2021 Ralf Kirchherr, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

from __future__ import annotations

import pathlib
from typing import Dict, List, Union

from rasa.shared.data import get_data_files, is_nlu_file
from rasa.shared.importers.utils import training_data_from_paths
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_END,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_VALUE,
    INTENT,
    TEXT,
)
from rasa.shared.nlu.training_data.formats.rasa import RasaReader, RasaWriter
from rasa.shared.nlu.training_data.formats.rasa_yaml import RasaYAMLWriter
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.utils.io import write_yaml

from nlubridge.nlu_dataset import EntityKeys, NluDataset


def from_rasa(
    filepath: Union[str, pathlib.Path], format: str = "yml", verbose=True
) -> NluDataset:
    """
    Load data stored in Rasa's yml- resp. json-format as NLUdataset.

    :param filepath: file path to read data from (Rasa specific yml- or json-format).
                     If 'yml' format is selected, this can be a directory with several
                     files.
    :param format: Input format, 'yml' (default) or 'json'
    :return: The loaded dataset as NLUdataset object
    """
    if format == "yml":
        training_files = get_data_files([filepath], is_nlu_file)
        # TODO: Replace with log messages!
        if verbose:
            print(f"Found {len(training_files)} NLU files")
            print("Reading data...")
        # Language tag in this function seems to be meaningless as in Rasa source code
        # apparently is never set to anything else than default "en"
        training_data = training_data_from_paths(training_files, "en")
    elif format == "json":
        training_data = RasaReader().read(filename=filepath)
    else:
        raise ValueError(f"Unknown format {format!r}")

    texts = []
    intents = []
    entities = []
    for message in training_data.training_examples:
        texts.append(message.get(TEXT))
        intents.append(message.get(INTENT))
        es = convert_entities_to_nludataset(message)
        entities.append(es)

    return NluDataset(texts, intents, entities)


def convert_entities_to_nludataset(message: Message) -> List[Dict]:
    """
    Convert entity dicts from Rasa format to NluDataset format.

    It will change the key names for entity type, start index and end index to those
    used in NluDataset and keep all additional keys from the Rasa formatted entity.

    :param message: A Rasa Message object
    :return: list of entities (dicts) in new format
    """
    es = []
    for e in message.get(ENTITIES, []):
        entity = {
            EntityKeys.TYPE: e.get(ENTITY_ATTRIBUTE_TYPE),
            EntityKeys.START: e.get(ENTITY_ATTRIBUTE_START),
            EntityKeys.END: e.get(ENTITY_ATTRIBUTE_END),
        }
        # Add any custom keys defined in the source structure
        for key in e.keys():
            if key not in [
                ENTITY_ATTRIBUTE_TYPE,
                ENTITY_ATTRIBUTE_START,
                ENTITY_ATTRIBUTE_END,
            ]:
                entity[key] = e[key]
        es.append(entity)
    return es


def to_rasa(
    dataset: NluDataset, filepath: Union[str, pathlib.Path], format: str = "yml"
):
    """
    Write dataset in Rasa's yml- or json-format.

    :param dataset: Dataset to be converted
    :param filepath: Path of the output yml file
    :param format: Output format, 'yml' (default) or 'json'
    """
    # TODO: This method should share code with the Rasa._convert() method
    messages = []
    for text, intent, entities in zip(dataset.texts, dataset.intents, dataset.entities):
        formatted_entities = []
        for e in entities:
            formatted_entity = {
                ENTITY_ATTRIBUTE_TYPE: e[EntityKeys.TYPE],
                ENTITY_ATTRIBUTE_START: e[EntityKeys.START],
                ENTITY_ATTRIBUTE_END: e[EntityKeys.END],
                # Please note: This sets just the default 'value' (if the input dataset
                # provides an explicit 'value' parameter, it will be adapted accordingly
                # in the section for custom keys below)
                ENTITY_ATTRIBUTE_VALUE: text[e[EntityKeys.START] : e[EntityKeys.END]],
            }
            # Add any custom keys defined in the source structure
            for key in e.keys():
                if key not in [EntityKeys.TYPE, EntityKeys.START, EntityKeys.END]:
                    formatted_entity[key] = e[key]
            formatted_entities.append(formatted_entity)
        example = {
            TEXT: text,
            INTENT: intent if intent is not None else "default_intent",
            ENTITIES: formatted_entities,
        }
        message = Message(data=example)
        messages.append(message)

    training_data = TrainingData(training_examples=messages)
    if format == "yml":
        mry = RasaYAMLWriter()
        md = mry.training_data_to_dict(training_data)
        write_yaml(md, filepath)
    elif format == "json":
        mrj = RasaWriter()
        mrj.dump(filepath, training_data)
    else:
        raise ValueError(f"Unsupported format {format!r}")



import uuid

from nlubridge.nlu_dataset import NluDataset


class ToyDataset(NluDataset):
    def __init__(self, **kw_args):
        """Create a NLUdataset with two samples and intent and entity annotations."""
        texts = [
            "Book me a flight from Cairo to Redmond next Thursday",
            "What's the weather like in Seattle?",
        ]
        intents = ["BookFlight", "GetWeather"]
        entities = [
            [
                {"entity": "LocationFrom", "start": 22, "end": 27},
                {"entity": "LocationTo", "start": 31, "end": 38},
            ],
            [{"entity": "Location", "start": 27, "end": 34}],
        ]
        super().__init__(texts, intents, entities, **kw_args)


class SyntheticDataset(NluDataset):
    def __init__(self, n_samples, intents, **kw_args):
        """
        Create an NLUdataset with n_samples samples.

        Utterances are random character sequences and intents from the passed intents
        are assigned to them evenly distributed.
        """
        texts = [str(uuid.uuid1()) for _ in range(n_samples)]
        intents = intents * (round(n_samples / len(intents)) + 1)
        intents = intents[:n_samples]
        super().__init__(texts, intents, None, **kw_args)


class TrainingDataset(NluDataset):
    def __init__(self):
        """Create a NLUdataset for testing with two intents and no entities."""
        help = [
            "I need help",
            "help me",
            "please help me",
            "what can i do?",
            "how does this work?",
            "help, i need somebody",
            "hello can i get help",
            "i really need help",
            "i dont know how this works",
            "how is this supposed to be used",
            "can someone help me",
            "how do i do this",
            "show me how this works please",
            "i need a manual",
            "can i read the manual",
            "instructions please",
            "i need better instructions",
            "i have trouble using this",
            "i cannot use this",
            "I wish i knew how this works" "",
        ]
        affirm = [
            "say your name please",
            "what's your name",
            "how can i call you",
            "how may i dub you",
            "can you tell me your name",
            "who are you",
            "please tell me your name",
            "i wish i knew your name",
            "how do you call yourself",
            "how do people call you",
        ]
        intents = ["help"] * 20 + ["affirm"] * 10
        texts = help + affirm
        super().__init__(texts, intents, None)



# Copyright (c) Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import logging
import pickle

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import SGDClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.svm import OneClassSVM

from ..nlu_dataset import OUT_OF_SCOPE_TOKEN
from .vendor import Vendor


logger = logging.getLogger(__name__)


class CharNgramIntentClassifier(Vendor):
    def __init__(self, anomaly_clf_nu=0):
        """
        Interface to a custom intent classifier based on character n-grams.

        Intent classification is based on character n-grams TFIDF
        model with hyper parameter tuning. In addition, a second
        classifier detects out-of-scope examples with an anomaly
        detection approach.

        :param anomaly_clf_nu: parameter that sets the
            aggressiveness of the anomaly detection (set to 0 to switch
            off)
        :type anomaly_clf_nu: float
        """
        self._alias = self.name
        self._clf = Model1(
            none_class=OUT_OF_SCOPE_TOKEN, verbose=False, anomaly_clf_nu=anomaly_clf_nu
        )

    def train_intent(self, dataset):  # noqa D102
        X = dataset.texts
        y = dataset.intents
        self._clf.fit(X, y)
        return self

    def test_intent(self, dataset, return_probs=False):  # noqa D102
        X = dataset.texts
        intents, probs = self._clf.predict(X, return_probs=True)
        intents = list(intents)
        probs = list(probs)
        if return_probs:
            return intents, probs
        return intents


class Model1:
    """
    Model1 implementation in original code.

    TF-IDF ==> Anomaly
    TF-IDF ==> SGD
    """

    def __init__(
        self, none_class="no_intent", verbose=False, anomaly_clf_nu=0.95
    ):  # noqa D102
        self.none_class_ = none_class
        self.verbose_ = verbose
        self.anomaly_clf_nu_ = anomaly_clf_nu
        self.models_path = "models/"
        self.anomaly_clf = Pipeline(
            [
                ("tfidf", TfidfVectorizer(max_df=1.0, min_df=2, analyzer="word")),
                ("svm", OneClassSVM(kernel="linear", nu=self.anomaly_clf_nu_)),
            ]
        )
        self.intent_clf = Pipeline(
            [
                (
                    "tfidf",
                    TfidfVectorizer(
                        min_df=1,
                        max_df=1.0,
                        ngram_range=(1, 4),
                        use_idf=True,
                        analyzer="char",
                    ),
                ),
                ("sgd", SGDClassifier(alpha=7e-06, max_iter=20, n_jobs=1, loss="log")),
            ]
        )

        self.intent_clf_param_space = {
            "tfidf__max_df": [0.5, 0.75, 1.0],
            "tfidf__ngram_range": [(1, 4), (1, 5), (1, 6)],
            "sgd__alpha": [1e-05, 1e-06, 1e-07],
        }

    def fit(self, X, y=None):  # noqa D102
        logger.debug("Training started")
        X_with_intent = [x for x, y in zip(X, y) if not y == self.none_class_]
        if self.anomaly_clf_nu_ > 0:
            self.trained_anomaly_clf = self.anomaly_clf.fit(X_with_intent)
        self.trained_intent_clf = self._hyper_fit(
            self.intent_clf, self.intent_clf_param_space, X, y
        )
        # self.save()
        logger.debug("Training finished")

    def predict(self, X, y=None, return_probs=False):  # noqa D102
        if self.anomaly_clf_nu_ > 0:
            y_has_intent = self.trained_anomaly_clf.predict(X)

        y_intent = self.trained_intent_clf.predict(X)
        probs = self.trained_intent_clf.predict_proba(X)
        probs = probs.max(axis=1)

        if self.anomaly_clf_nu_ > 0:
            intents = [
                y if has_intent < 0 else self.none_class_
                for y, has_intent in zip(y_intent, y_has_intent)
            ]
        else:
            intents = y_intent

        if return_probs:
            return intents, probs

        return intents

    def _hyper_fit(self, classifier, param_space, X, y):  # noqa D102
        search = GridSearchCV(
            classifier, param_space, n_jobs=-1, cv=5, verbose=self.verbose_, refit=True
        )
        # t = time()
        search.fit(X, y)

        # TODO: This code snippet prints some diagnostic info re the hyperparameter
        #  search. We may want to reactivate it, but then maybe output to logger?
        # if self.verbose_:
        #     pprint(param_space)
        #     print("grid search time for ", time() - t, "s")
        #     print("Best score: %0.3f" % search.best_score_)
        #     print("Best parameters set:")
        #     best_parameters = search.best_params_
        #     for param_name in sorted(param_space.keys()):
        #         print("\t%s: %r" % (param_name, best_parameters[param_name]))

        return search

    def load(self, filename="model1"):  # noqa D102
        f = open(self.models_path + filename, "rb")
        tmp_dict = pickle.load(f)
        f.close()

        self.__dict__.update(tmp_dict)

    def save(self, filename="model1"):  # noqa D102
        f = open(self.models_path + filename, "wb")
        pickle.dump(self.__dict__, f, 2)
        f.close()



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

from __future__ import annotations

import os
import pathlib
from typing import List, Optional, Tuple, Union

from rasa.nlu import config
from rasa.nlu.model import Trainer
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_END,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_VALUE,
    INTENT,
    INTENT_NAME_KEY,
    INTENT_RANKING_KEY,
    PREDICTED_CONFIDENCE_KEY,
    TEXT,
)
from rasa.shared.nlu.training_data.formats.rasa import RasaReader
from rasa.shared.nlu.training_data.training_data import TrainingData

from nlubridge.dataloaders.rasa import convert_entities_to_nludataset
from nlubridge.nlu_dataset import EntityKeys, NBestKeys, NluDataset
from nlubridge.vendors import Vendor


DEFAULT_INTENT_RASA_CONFIG_PATH = os.path.join(
    pathlib.Path(__file__).parent.absolute(), "config", "rasa_nlu_config.yml"
)
ENTITY_KEY_VALUE = "value"  # Rasa provides an explicit value parameter for its entities


class Rasa2(Vendor):
    def __init__(self, model_config: Optional[str] = None):
        """
        Interface for the `Rasa NLU <https://github.com/RasaHQ/rasa>`_.

        Uses the default pipeline as of January 22 2021. No algorithmic
        details have been researched. The configuration file can be
        found in the config directory within this directory. A custom
        pipeline can be provided as argument to the class constructor.

        :param model_config: filepath to a Rasa config file
        """
        self._alias = self.name
        self._config = model_config
        self._interpreter = None

    def train(self, dataset: NluDataset) -> Rasa2:
        """
        Train intent and/or entity classification.

        :param dataset: Training data
        :return: It's own Rasa object
        """
        self._config = self._config if self._config else DEFAULT_INTENT_RASA_CONFIG_PATH
        training_data = self._convert(dataset)
        trainer = Trainer(config.load(self._config))
        self._interpreter = trainer.train(training_data)
        return self

    def train_intent(self, dataset: NluDataset) -> Rasa2:
        """
        Train intent classification.

        This method is mainly for compatibility reasons, as it in case of Rasa identical
        to the `train` method.

        :param dataset: Training data
        :return: It's own Rasa object
        """
        return self.train(dataset)

    def test(self, dataset: NluDataset) -> NluDataset:
        """
        Test a given dataset.

        Test a given dataset and obtain the intent and/or entity classification results
        in the NLUdataset format.

        :param dataset: Input dataset to be tested
        :return: NLUdataset object comprising the classification results. The list of
            the predicted intent classification probabilities are accessible via the
            additional attribute 'probs' (List[float]).
        """
        intents: List[str] = []
        n_best_lists: List[List[dict]] = []
        entities_list: List[List[dict]] = []
        if self._interpreter is None:
            raise Exception("Rasa2 classifier has to be trained first!")
        for text in dataset.texts:
            result = self._interpreter.parse(text)
            intent = result.get(INTENT, {}).get(INTENT_NAME_KEY)
            entities = convert_entities_to_nludataset(result)
            nbest = [
                {
                    NBestKeys.INTENT: ranked.get(INTENT_NAME_KEY),
                    NBestKeys.CONFIDENCE: ranked.get(PREDICTED_CONFIDENCE_KEY),
                }
                for ranked in result.get(INTENT_RANKING_KEY, [])
            ]

            intents.append(intent)
            n_best_lists.append(nbest)
            entities_list.append(entities)

        res = NluDataset(dataset.texts, intents, entities_list, n_best_lists)
        return res

    def test_intent(
        self, dataset: NluDataset, return_probs: bool = False
    ) -> Union[List[str], Tuple[List[str], List[float]]]:
        """
        Test a given dataset and obtain just the intent classification results.

        :param dataset: The dataset to be tested
        :param return_probs: Specifies if the probability values should be returned
            (default is False)
        :return: Either a list of predicted intent classification results or a tuple of
            predicted intent classification and probabilites results (depeding on
            argument 'return_probs')
        """
        intents: List[str] = []
        probs: List[float] = []
        if self._interpreter is None:
            raise Exception("Rasa2 classifier has to be trained first!")
        for text in dataset.texts:
            result = self._interpreter.parse(text)
            intent = result.get(INTENT, {}).get(INTENT_NAME_KEY)
            prob = result.get(INTENT, {}).get(PREDICTED_CONFIDENCE_KEY)
            intents.append(intent)
            probs.append(prob)
        if return_probs:
            return intents, probs
        return intents

    @staticmethod
    def _convert(dataset: NluDataset) -> TrainingData:
        """
        Convert a NLUdataset to a Rasa TrainingData object.

        :param dataset: NLUdataset to be converted
        :return: Rasa TrainingData object
        """
        examples = []

        for text, intent, entities in zip(
            dataset.texts, dataset.intents, dataset.entities
        ):
            example = {
                TEXT: text,
                INTENT: intent if intent is not None else "default_intent",
                ENTITIES: [],
            }
            for entity in entities:
                formatted_entity = {
                    ENTITY_ATTRIBUTE_TYPE: entity[EntityKeys.TYPE],
                    ENTITY_ATTRIBUTE_START: entity[EntityKeys.START],
                    ENTITY_ATTRIBUTE_END: entity[EntityKeys.END],
                    # Please note: This sets just the default 'value' (if the input
                    # dataset provides an explicit 'value' parameter, it will be adapted
                    # accordingly in the section for custom keys below)
                    ENTITY_ATTRIBUTE_VALUE: text[
                        entity[EntityKeys.START] : entity[EntityKeys.END]
                    ],
                }
                # Add any custom keys defined in the source structure
                for key in entity.keys():
                    if key not in [EntityKeys.TYPE, EntityKeys.START, EntityKeys.END]:
                        formatted_entity[key] = entity[key]
                example[ENTITIES].append(formatted_entity)  # type: ignore[attr-defined]
            examples.append(example)

        training_data = {
            "rasa_nlu_data": {
                "common_examples": examples,
                "regex_features": [],
                "entity_synonyms": [],
            }
        }

        return RasaReader().read_from_json(training_data)



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

from nlubridge.nlu_dataset import NluDataset

from .utils import from_json


def from_luis(filepath) -> NluDataset:
    """
    Load data in LUIS format as NLUdataset.

    :param filepath: file path to the LUIS-formatted data.
    :type filepath: str
    """
    dataset = from_json(
        path=filepath,
        text_key="text",
        intent_key="intentName",
        entities_key="entityLabels",
        entity_type_key="entityName",
        entity_start_key="startCharIndex",
        entity_end_key="endCharIndex",
        end_index_add_1=True,
    )
    return dataset



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT
"""Implementations of interfaces to different vendors."""

import sys
from typing import TYPE_CHECKING

from lazy_imports import LazyImporter

from .vendor import Vendor  # noqa: 401


if TYPE_CHECKING:
    from .char_ngram_intent_classifier import CharNgramIntentClassifier  # noqa: F401
    from .fasttext import FastText  # noqa: F401
    from .luis import Luis  # noqa: F401
    from .rasa2 import Rasa2  # noqa: F401
    from .rasa3 import Rasa3  # noqa: F401
    from .spacy import Spacy  # noqa: F401
    from .tfidf_intent_classifier import TfidfIntentClassifier  # noqa: 401
    from .vendor import Vendor  # noqa: 401
    from .watson_assistant import WatsonAssistant  # noqa: F401
else:
    _import_structure = {
        "fasttext": ["FastText"],
        "luis": ["Luis"],
        "rasa2": ["Rasa2"],
        "rasa3": ["Rasa3"],
        "spacy": ["Spacy"],
        "char_ngram_intent_classifier": ["CharNgramIntentClassifier"],
        "tfidf_intent_classifier": ["TfidfIntentClassifier"],
        "vendor": ["Vendor"],
        "watson_assistant": ["WatsonAssistant"],
    }
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
        # extra_objects={"__version__": __version__},
    )



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import logging

from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer
from sklearn.linear_model import SGDClassifier
from sklearn.pipeline import Pipeline

from .vendor import Vendor


logger = logging.getLogger(__name__)


class TfidfIntentClassifier(Vendor):
    def __init__(self):
        """
        Interface to a TFIDF text classifier.

        Simple word based tf-idf plus a SVM classifier. It can be used
        as a baseline and to quickly tests a script by providing a fast
        algorithm.

        Pros:
            * Very fast classifier
            * Works relatively well for most problems
            * Provides a good baseline that custom models should beat

        Cons:
            * Can't use external information (ex: pretrained word
              embeddings)
        """
        self._alias = self.name
        self._clf = Pipeline(
            [
                ("vect", CountVectorizer()),
                ("tfidf", TfidfTransformer()),
                (
                    "clf",
                    SGDClassifier(
                        loss="log",
                        penalty="l2",
                        alpha=1e-5,
                        random_state=42,
                        max_iter=700,
                        tol=None,
                        n_jobs=-1,
                    ),
                ),
            ]
        )

    def train_intent(self, dataset):
        """Train intent classifier."""
        logger.info(f"Training on {dataset.n_samples} samples")
        X = dataset.texts
        y = dataset.intents
        self._clf.fit(X, y)
        return self

    def test_intent(self, dataset, return_probs=False):
        """Test intent classifier."""
        logger.info(f"Testing on {dataset.n_samples} samples")
        X = dataset.texts
        probs = self._clf.predict_proba(X)
        pred_idxs = probs.argmax(axis=1)
        winner_class_probs = probs.max(axis=1)
        winner_class_probs = list(winner_class_probs)
        intents = self._clf.classes_[pred_idxs]
        intents = list(intents)

        if return_probs:
            return intents, winner_class_probs
        return intents



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT
"""Provides a unified API to several popular intent recognition applications."""  # noqa D400

import sys
from typing import TYPE_CHECKING

from lazy_imports import LazyImporter


__version__ = "1.0.2"


if TYPE_CHECKING:
    from .dataloaders.huggingface import from_huggingface  # noqa: F401
    from .dataloaders.luis import from_luis  # noqa: F401
    from .dataloaders.rasa import from_rasa, to_rasa  # noqa: F401, F811
    from .dataloaders.utils import from_csv, from_json  # noqa: F401
    from .dataloaders.watson_assistant import from_watson_assistant  # noqa: F401
    from .nlu_dataset import (  # noqa: F401
        OUT_OF_SCOPE_TOKEN,
        EntityKeys,
        NBestKeys,
        NluDataset,
        concat,
    )
    from .vendors.vendor import Vendor  # noqa: 401

else:
    _import_structure = {
        "dataloaders.huggingface": ["from_huggingface"],
        "dataloaders.luis": ["from_luis"],
        "dataloaders.rasa": ["from_rasa", "to_rasa"],
        "dataloaders.utils": ["from_json", "from_csv"],
        "dataloaders.watson_assistant": ["from_watson_assistant"],
        "nlu_dataset": [
            "OUT_OF_SCOPE_TOKEN",
            "EntityKeys",
            "NBestKeys",
            "NluDataset",
            "concat",
        ],
        "vendors.vendor": ["Vendor"],
    }

    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
        # extra_objects={"__version__": __version__},
    )



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import collections
import csv
import json
from pathlib import Path
from typing import Dict, List, Optional, Union

from nlubridge.nlu_dataset import EntityKeys, NluDataset


def from_csv(
    path: Union[Path, str], text_col: Union[str, int], intent_col: Union[str, int]
) -> NluDataset:
    """
    Load dataset (only text and intents) from a csv file.

    text_col and intent_col can be either integer indices of the respective columns
    (start index is 0), or, if the first row of the file holds the column names, the
    repective column name strings.
    """
    columns = collections.defaultdict(list)

    with open(path) as f:
        reader = csv.DictReader(f)
        for row in reader:
            for k, v in row.items():
                columns[k].append(v)

    if isinstance(text_col, int):
        key = list(columns.keys())[text_col]
        texts = [key, *columns[key]]
    else:
        texts = columns[text_col]

    if isinstance(intent_col, int):
        key = list(columns.keys())[intent_col]
        intents = [key, *columns[key]]
    else:
        intents = columns[intent_col]

    ds = NluDataset(texts, intents)
    return ds


def _convert_entities_for_nludataset(
    entities, type_key, start_key, end_key, end_index_add_1: bool
):
    ex_entities = []
    for entity in entities:
        formatted_entity = {
            EntityKeys.TYPE: entity[type_key],
            EntityKeys.START: entity[start_key],
            EntityKeys.END: entity[end_key] + end_index_add_1,
        }
        # Add any custom keys defined in the source json
        for key in entity.keys():
            if key not in [type_key, start_key, end_key]:
                formatted_entity[key] = entity[key]
        ex_entities.append(formatted_entity)
    return ex_entities


def from_json(
    path: Optional[Union[Path, str]] = None,
    examples: Optional[List[Dict]] = None,
    text_key: str = "text",
    intent_key: str = "intent",
    entities_key: str = "entities",
    entity_type_key: str = "entity",
    entity_start_key: str = "start",
    entity_end_key: str = "end",
    end_index_add_1: bool = False,
):
    """
    Load the dataset form a json string.

    The json string should have the structure in the TestDataset class.
    """
    if (not path and not examples) or (path and examples):
        raise ValueError("Exactly one of path or examples arguments needs to be given")

    if path:
        with open(path, "r") as f:
            examples = json.load(f)

    texts, intents, entities = [], [], []
    for example in examples:  # type: ignore[union-attr]
        texts.append(example[text_key])
        intents.append(example[intent_key])
        example_entities = _convert_entities_for_nludataset(
            example[entities_key],
            entity_type_key,
            entity_start_key,
            entity_end_key,
            end_index_add_1,
        )
        entities.append(example_entities)
    dataset = NluDataset(texts, intents, entities)
    return dataset



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT
"""Provides a unified API to several training data formats."""  # noqa D400



import copy
import os
from collections import Counter

import pytest
from sklearn.model_selection import KFold
from testing_data import SyntheticDataset, ToyDataset

from nlubridge import EntityKeys, NBestKeys, NluDataset, concat, from_json


FIXTURE_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "fixtures")


texts = [
    "Book me a flight from Cairo to Redmond next Thursday",
    "What's the weather like in Seattle?",
    "What's the weather in Berlin?",
]
intents = ["BookFlight", "GetWeather", "GetWeather"]
entities = [
    [
        {EntityKeys.TYPE: "Location::From", EntityKeys.START: 22, EntityKeys.END: 27},
        {EntityKeys.TYPE: "Location::To", EntityKeys.START: 31, EntityKeys.END: 38},
    ],
    [{EntityKeys.TYPE: "Location", EntityKeys.START: 27, EntityKeys.END: 34}],
    [{EntityKeys.TYPE: "Location", EntityKeys.START: 22, EntityKeys.END: 28}],
]
n_best_intents = [
    [
        {NBestKeys.INTENT: "BookFlight", NBestKeys.CONFIDENCE: 0.5},
        {NBestKeys.INTENT: "GetWeather", NBestKeys.CONFIDENCE: 0.4},
    ],
    [
        {NBestKeys.INTENT: "GetWeather", NBestKeys.CONFIDENCE: 0.9},
        {NBestKeys.INTENT: "BookFlight", NBestKeys.CONFIDENCE: 0.1},
    ],
    [
        {NBestKeys.INTENT: "BookFlight", NBestKeys.CONFIDENCE: 0.8},
        {NBestKeys.INTENT: "GetWeather", NBestKeys.CONFIDENCE: 0.2},
    ],
]


def test_init_ds_with_texts_intents_entities():
    ds = NluDataset(texts, intents, entities)
    assert ds.texts == texts
    assert ds.intents == intents
    assert ds.unique_intents == ["BookFlight", "GetWeather"]
    assert ds.n_intents == 2
    assert ds.entities == entities
    assert ds.unique_entities == ["Location", "Location::From", "Location::To"]
    assert ds.n_entities == 3


def test_init_from_data():
    _ds = NluDataset(texts, intents, entities)
    ds = NluDataset._from_data(_ds._data)
    assert ds.texts == texts
    assert ds.intents == intents
    assert ds.unique_intents == ["BookFlight", "GetWeather"]
    assert ds.n_intents == 2
    assert ds.entities == entities
    assert ds.unique_entities == ["Location", "Location::From", "Location::To"]
    assert ds.n_entities == 3


def test_crop_intent_names():
    ds = NluDataset(texts, intents, entities, max_intent_length=4)
    assert ds.texts == texts
    assert ds.intents == ["Book", "GetW", "GetW"]
    assert ds.unique_intents == ["Book", "GetW"]
    assert ds.n_intents == 2
    # check a warning is thrown when cropping makes intent names ambiguous
    with pytest.warns(UserWarning):
        ds2 = NluDataset(
            texts + ["bla"],
            intents + ["GetWhatever"],
            entities + [[]],
            max_intent_length=4,
        )
    assert ds2.n_intents == 2


def test_init_ds_without_intents_and_entities():
    ds = NluDataset(texts)
    assert len(ds.texts) == 3
    assert len(ds.intents) == 3
    assert ds.unique_intents == []
    assert ds.intents == [None, None, None]
    assert ds.n_intents == 0
    assert ds.intent_frequencies == Counter({None: 3})
    assert len(ds.entities) == 3
    assert ds.unique_entities == []
    assert ds.entities == [[], [], []]
    assert ds.n_entities == 0
    assert ds.confidences == [None, None, None]
    assert ds.n_best_intents == [[], [], []]


def test_init_with_nbest_list():
    ds = NluDataset(texts, intents, entities, n_best_intents)
    assert len(ds.n_best_intents) == 3
    assert ds.confidences == [0.5, 0.9, 0.8]
    assert ds._data[0][3] == n_best_intents[0]


def test_clip_by_intent_frequency():
    pass


def test__filter_by_index_list():
    pass


def test_ds_unique():
    ds = ToyDataset()
    assert ds.n_intents == 2
    assert ds.n_entities == 3


def test_ds_slicing():
    ds = NluDataset(texts, intents, entities)
    ds_sliced = ds[:2]
    assert isinstance(ds_sliced, NluDataset)
    assert len(ds_sliced) == 2
    ds_sliced = ds[2]
    assert isinstance(ds_sliced, NluDataset)
    assert len(ds_sliced) == 1


def test_ds_slicing_when_created_with_just_texts():
    # Ensure that when we pass a list of Nones for intents (during
    # construction of returned dataset in slicing) the returned
    # dataset is still valid
    ds = NluDataset(texts)
    sliced = ds[:2]
    assert len(sliced.texts) == 2
    assert len(sliced.intents) == 2
    assert len(sliced.entities) == 2
    assert len(sliced) == 2


def test_cross_validation_splits():
    ds = SyntheticDataset(10, intents=["intent1", "intent2"])
    for ds_train, ds_test in ds.cross_validation_splits():
        assert isinstance(ds_train, NluDataset)
        assert isinstance(ds_test, NluDataset)
        assert len(ds_train.texts) == 8
        assert len(ds_test.texts) == 2
        assert len(ds_train.texts) == len(ds_train.intents)
        assert len(ds_test.texts) == len(ds_test.intents)

    # Also test for leave-1-out (special case because the dataset has only a single
    # record)
    kf = KFold(n_splits=10)
    for ds_train, ds_test in ds.cross_validation_splits(kf):
        assert isinstance(ds_train, NluDataset)
        assert isinstance(ds_test, NluDataset)
        assert len(ds_train.texts) == 9
        assert len(ds_test.texts) == 1
        assert len(ds_train.texts) == len(ds_train.intents)
        assert len(ds_test.texts) == len(ds_test.intents)


def test_ds_from_joined():
    ds1 = ToyDataset()
    ds2 = ToyDataset()
    with pytest.warns(DeprecationWarning):
        joined = NluDataset.from_joined(ds1, ds2)
    assert joined.n_samples == ds1.n_samples + ds2.n_samples


def test_ds_join():
    ds1 = ToyDataset()
    ds2 = ToyDataset()
    ds1_samples = ds1.n_samples
    ds2_samples = ds2.n_samples
    joined = ds1.join(ds2)
    assert joined.n_samples == ds1_samples + ds2_samples
    # assert that the original dataset was not changed
    assert ds1_samples == ds1.n_samples
    joined2 = ds2.join(ds1)
    assert joined2.n_samples == ds1.n_samples + ds2.n_samples


def test_nlubridge_concat():
    ds1 = ToyDataset()
    ds2 = ToyDataset()
    ds3 = ToyDataset()
    ds1_samples = ds1.n_samples
    ds2_samples = ds2.n_samples
    ds3_samples = ds3.n_samples
    joined = concat([ds1, ds2, ds3])
    assert joined.n_samples == ds1_samples + ds2_samples + ds3_samples
    assert ds1_samples == ds1.n_samples
    assert ds2_samples == ds2.n_samples
    assert ds3_samples == ds3.n_samples


def test_to_records():
    ds = ToyDataset()
    with pytest.warns(DeprecationWarning):
        examples = ds.to_records()
    assert len(examples) == ds.n_samples


def test_to_json():
    ds = ToyDataset()
    examples = ds.to_json()
    assert isinstance(examples, list)
    assert isinstance(examples[0], dict)
    assert len(examples) == ds.n_samples


def test_to_json_from_json():
    ds = ToyDataset()
    examples = ds.to_json()
    ds2 = from_json(examples=examples)
    assert ds.texts == ds2.texts
    assert ds.intents == ds2.intents
    assert ds.entities == ds2.entities


def test_filter_by_intent_name():
    intents = [
        "intent_1",
        "intent_2",
        "intent_3",
        "intent_4",
        "intent_5",
        "intent_6",
    ]
    ds = SyntheticDataset(n_samples=100, intents=intents)
    excluded = ["intent_2", "intent_4"]
    filtered = ds.filter_by_intent_name(excluded=excluded)
    expected = [
        "intent_1",
        "intent_3",
        "intent_5",
        "intent_6",
    ]
    assert filtered.unique_intents == expected


# def test_select_first_n_intents():
#     # ds = select_first_n_intents(self, 2)
#     pass


def get_subsampling_ds():
    intents = (
        ["intent_1"] * 3000
        + ["intent_2"] * 3000
        + ["intent_3"] * 4500
        + ["intent_4"] * 500
    )
    ds = SyntheticDataset(n_samples=len(intents), intents=intents)
    return ds


def test_subsample_by_intent_frequency():
    ds = get_subsampling_ds()
    ds_sub = ds.subsample_by_intent_frequency(0.5, 1500)
    counter = Counter(ds_sub.intents)
    assert counter["intent_1"] == 2250
    assert counter["intent_2"] == 2250
    assert counter["intent_3"] == 3000
    assert counter["intent_4"] == 500
    # ensure intents and texts are not mixed up
    pairs = set(zip(ds.texts, ds.intents))
    pairs_sub = set(zip(ds_sub.texts, ds_sub.intents))
    assert len(pairs_sub - pairs) == 0


def test_subsample_by_intent_frequency_rate_is_zero():
    ds = get_subsampling_ds()
    ds_sub = ds.subsample_by_intent_frequency(0, 1500)
    counter = Counter(ds_sub.intents)
    assert counter["intent_1"] == 1500
    assert counter["intent_2"] == 1500
    assert counter["intent_3"] == 1500
    assert counter["intent_4"] == 500


def test_subsample_by_intent_frequency_rate_is_one():
    ds = get_subsampling_ds()
    ds_sub = ds.subsample_by_intent_frequency(1, 1500)
    counter = Counter(ds_sub.intents)
    assert counter["intent_1"] == 3000
    assert counter["intent_2"] == 3000
    assert counter["intent_3"] == 4500
    assert counter["intent_4"] == 500


def test_shuffle():
    ds = SyntheticDataset(n_samples=10, intents=["a", "b", "c"])
    shuffled = copy.deepcopy(ds).shuffle()
    assert ds.n_samples == shuffled.n_samples
    assert set(ds.texts) == set(shuffled.texts)
    assert ds.texts != shuffled.texts


def test_train_test_split():
    ds = SyntheticDataset(n_samples=10, intents=["a", "b", "c"])
    train, test = ds.train_test_split(test_size=0.3)
    assert train.n_samples == 7
    assert test.n_samples == 3


def test_cv_splits_texts_are_list():
    # addresses a former error where in the cv method tuples were
    # passed and to the dataset constructor for the partitioning
    ds = SyntheticDataset(n_samples=20, intents=["a", "b", "c"])
    for train_ds, test_ds in ds.cross_validation_splits():
        assert isinstance(train_ds.texts, list)
        assert isinstance(train_ds.intents, list)
        assert isinstance(train_ds.entities, list)
        assert isinstance(test_ds.texts, list)


def test_sample():
    size = 0.1
    ds = SyntheticDataset(n_samples=2000, intents=["a", "b", "c"])
    sampled = ds.sample(size=size)
    assert ds.n_intents == sampled.n_intents

    ratio = sampled.n_samples / ds.n_samples
    tolerance = 0.001
    assert (ratio - size) < tolerance

    # test reproduciblity
    ds1 = ds.sample(1000, random_state=42)
    ds2 = ds.sample(1000, random_state=42)
    assert ds1.texts[0] == ds2.texts[0]

    # ds3 = ds1.sample(100)
    # assert ds1.texts[0] != ds3.texts[0]



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import concurrent.futures
import datetime
import json
import logging
import os
import sys
import time

import requests
from ratelimit import rate_limited

from ..nlu_dataset import OUT_OF_SCOPE_TOKEN
from .vendor import Vendor


sys.path.append("../..")
logger = logging.getLogger(__name__)
global n_calls
global start_time


def _unwrap_self(arg, **kwarg):
    return Luis.test_single_intent(*arg, **kwarg)


class Luis(Vendor):
    AUTHORING_RATE_LIMIT = 4.99  # queries per second
    SUBSCRIPTION_RATE_LIMIT = 4.9  # queries per second

    intent_trainable = True
    entity_trainable = True

    def __init__(
        self,
        authoring_key=None,
        subscription_key=None,
        app_id=None,
        endpoint=None,
        version="0.1",
    ):
        """Interface for Microsoft LUIS."""
        self._alias = self.name
        endpoint = endpoint or os.getenv("LUIS_ENDPOINT")
        if endpoint is None:
            ValueError(
                "endpoint not passed and not found under environment variable"
                "LUIS_ENDPOINT"
            )
        authoring_key = authoring_key or os.getenv("LUIS_AUTHORING_KEY")
        if authoring_key is None:
            ValueError(
                "authoring_key not passed and not found under environment "
                "variable LUIS_AUTHORING_KEY"
            )
        subscription_key = subscription_key or os.getenv("LUIS_SUBSCRIPTION_KEY")
        if subscription_key is None:
            ValueError(
                "subscription_key not passed and not found under environment "
                "variable LUIS_SUBSCRIPTION_KEY"
            )
        app_id = app_id or os.getenv("LUIS_APP_ID")
        if app_id is None:
            ValueError(
                "app_id not passed and not found under environment variable "
                "LUIS_APP_ID"
            )
        self._endpoint = endpoint
        self._subscription_key = subscription_key
        self._app_id = app_id
        self._version = version
        self._session = requests.Session()
        self._session.headers.update({"Ocp-Apim-Subscription-Key": authoring_key})
        logger.debug(f"Created new app with id {self._app_id}")

    @property  # type: ignore
    @rate_limited(AUTHORING_RATE_LIMIT)
    def requests(self):
        """Util method to access self.session."""
        return self._session

    def _get_base_url(self, cmd, add_version=True):
        base_url = requests.compat.urljoin(self._endpoint, self._app_id)
        base_url += "/"
        if add_version:
            base_url += f"versions/{self._version}/"
        base_url += f"{cmd}"
        return base_url

    def _convert(self, dataset):
        """
        Convert dataset to LUIS format.

        This function accepts a dataset as input and returns a list of samples
        in the vendor specific expected format.
        """
        examples = []

        for text, intent, entities in dataset:
            example = {
                "text": text,
                "intentName": intent,
                "entityLabels": [
                    {
                        "entityName": entity["entity"],
                        "startCharIndex": entity["start"],
                        "endCharIndex": entity["end"],
                    }
                    for entity in entities
                ],
            }  # yapf: disable
            examples.append(example)

        return json.dumps(examples)

    def _add_application(self, name=None, description=None):
        """Create a new LUIS app."""
        timestamp = "{:%Y-%m-%d_%H:%M:%S}".format(datetime.datetime.now())
        body = {
            "name": name or f"test_app_{timestamp}",
            "description": description or "",
            "culture": "de-de",
            "usageScenario": "",
            "domain": "",
            "initialVersionId": self._version,
        }

        response = self.requests.post(self._endpoint, data=body)
        app_id = response.json()
        return app_id

    def _delete_application(self, app_id=None):
        """
        Delete an application.

        If an `app_id` is given that app will be deleted.
        Otherwise `self.app_id
        """
        if app_id is None:
            app_id = self._app_id

        url = requests.compat.urljoin(self._endpoint, app_id)
        response = self.requests.delete(url)
        return response.json()

    def _clone_version(self, old_version="0.1", new_version="perf_test"):
        """
        Clone a version of the application.

        The version '0.1' is kept empty and used as a blank state.
        Then cloned versions are used for the tests.
        """
        self._version = old_version
        url = self._get_base_url("clone", add_version=True)
        body = {"version": new_version}
        response = self.requests.post(url, data=body)
        response = response.json()

        if response == new_version:
            # clone created successfully
            self._version = new_version
            logger.debug(f"Created new version {response}")

        return response

    def _delete_app_version(self, version="perf_test"):
        """Delete an application version."""
        if version is None:
            version = self._version

        if version == "0.1":
            raise Exception("You cant delete the main version")
        self._version = "perf_test"
        url = self._get_base_url("", add_version=True)

        return self._session.delete(url)

    def _create_intents(self, dataset):
        """Add intent classifiers to the application."""
        if dataset.n_intents > 500:
            raise ValueError(
                "LUIS does not support more than 500 intents:\n"
                "https://docs.microsoft.com/en-us/azure/cognitive-services/luis/luis-boundaries"  # noqa: E501
            )

        url = self._get_base_url("intents")
        for intent in dataset.unique_intents:
            body = {"name": intent}
            response = self.requests.post(url, data=body)
            response = response.json()
            logger.debug(f"{response}")

    def _create_entities(self, dataset):
        """Add entity extractors to the application."""
        if dataset.n_entities > 30:
            raise ValueError(
                "LUIS does not support more than 30 entities.\n"
                "https://docs.microsoft.com/en-us/azure/cognitive-services/luis/luis-boundaries"  # noqa: E501
            )

        url = self._get_base_url("entities")
        for entity in dataset.unique_entities:
            body = {"name": entity}
            response = self.requests.post(url, data=body)
            response = response.json()
            logger.debug(f"{response}")

    def _upload_samples(self, dataset):
        logger.debug(f"Uploading samples for dataset {dataset.name}")
        logger.debug(f"Number of samples {dataset.n_samples}")

        logger.debug("Creating intents ...")
        self._create_intents(dataset)
        logger.debug("Creating entities ...")
        self._create_entities(dataset)

        url = self._get_base_url("examples")
        logger.debug("Uploading examples ...")

        # the /examples endpoint only accepts 100 length batches
        step = 100
        for start in range(0, dataset.n_samples, step):
            end = start + step
            idx = slice(start, end)
            ds = dataset[idx]
            body = self._convert(ds)
            response = self.requests.post(url, data=body)
            n_uploaded = len(response.json())
            logger.debug(f"Batch with {n_uploaded} samples uploaded")

        return response.json()

    def _train(self):
        url = self._get_base_url("train", add_version=True)
        logger.debug("Launching training ...")
        response = self.requests.post(url)

        while not self._is_trained:
            logger.debug("Waiting for the model to finish training")
            time.sleep(0.5)
        response = self._publish(staging=True)
        logger.debug("Finished publishing with response %s", response.json())
        return response

    def _get_train_status(self):
        url = self._get_base_url("train", add_version=True)
        response = self.requests.get(url)
        return response.json()

    @property
    def _is_trained(self):
        status = self._get_train_status()
        # logger.debug(f'Training status debug details: {status}')
        status_details = [
            each.get("details", {}).get("status", "") == "Success" for each in status
        ]
        logger.debug(f"Training status details: {status_details}")
        return all(status_details)  # yapf: disable

    def _is_published(self):
        url = self._get_base_url("", add_version=False)
        response = self.requests.get(url)
        logger.debug("Checked publishing with response %s", response.json())
        return response.json()

    def _publish(self, staging=True):
        """Publish a specific version of the application."""
        url = self._get_base_url("publish", add_version=False)
        body = {"versionId": self._version, "isStaging": staging}
        response = self.requests.post(url, data=body)
        return response

    def train_intent(self, dataset):
        """Train intent classifier."""
        logger.info(f"Training on {dataset.n_samples} samples")
        self._delete_app_version(version="perf_test")
        self._clone_version(old_version="0.1", new_version="perf_test")
        self._upload_samples(dataset)

        # not clear if removing "response = "could have side effects
        response = self._train()  # noqa: F841

        logger.debug("Will now start measuring time and calls...")
        global n_calls
        global start_time
        n_calls = 0
        start_time = time.time()
        return self

    @rate_limited(SUBSCRIPTION_RATE_LIMIT)
    def test_single_intent(self, query, return_probs=False):
        """
        Predict the intent for an utterance.

        LUIS doesn't provide a batch testing API so
        we have to test one intent at a time.
        """
        url = self._get_base_url(cmd="", add_version=False)
        url = url.replace("/api", "")
        # utterances have a limit of max 500 characters
        # https://docs.microsoft.com/en-us/azure/cognitive-services/luis/luis-boundaries
        query = query[:500]
        body = {
            "q": query,
            "staging": "true",
            "log": False,
            "subscription-key": self._subscription_key,
        }
        # here we dont use the session because the auth key headers
        # take preference over the subscription key
        response = requests.get(url, params=body)

        # logger.debug(f'Full url: {response.request.url}')

        response = response.json()
        intent = response.get("topScoringIntent", {}).get("intent")
        prob = response.get("topScoringIntent", {}).get("score", 0)

        if intent is None:
            intent = OUT_OF_SCOPE_TOKEN

        global n_calls
        global start_time

        n_calls += 1
        if n_calls % 50 == 0:
            logger.debug(f"Total calls: {n_calls}")
            calls_per_second = n_calls / (time.time() - start_time + 0.0001)
            logger.debug(f"Calls per second: {calls_per_second}")

        if return_probs:
            return intent, prob
        return intent

    def test_intent(self, dataset, return_probs=False):
        """Test intent classifier."""
        logger.info(f"Testing on {dataset.n_samples} samples")
        selfs = [self] * dataset.n_samples
        _return_probs = [return_probs] * dataset.n_samples
        tups = list(zip(selfs, dataset.texts, _return_probs))

        with concurrent.futures.ThreadPoolExecutor(max_workers=15) as executor:
            intents = list(executor.map(_unwrap_self, tups))

        if return_probs:
            intents, probs = list(zip(*intents))
            return intents, probs
        return intents



from dotenv import load_dotenv
from test_vendors import (
    assert_multiple_utterances_predicted,
    assert_preds_are_intents,
    assert_return_probs,
)

from nlubridge import NluDataset


def test_watson(train_data):
    from nlubridge.vendors.watson_assistant import WatsonAssistant

    # load environment variables so Watson uses them
    load_dotenv()

    # test initialization
    watson = WatsonAssistant()

    # test train_intent()
    watson.train_intent(train_data)

    # test test_intent() (bulk_classify)
    assert_preds_are_intents(watson, train_data.unique_intents)
    assert_return_probs(watson, train_data.unique_intents)
    assert_multiple_utterances_predicted(watson, train_data)
    # test handling of rate exceeded (only 250/min allowed)
    cases90 = train_data + train_data + train_data
    cases360 = cases90 + cases90 + cases90 + cases90
    preds, probs = watson.test_intent(cases360, return_probs=True)
    assert len(preds) == 360
    assert len(probs) == 360
    # TODO: fix vendor for test: assert_oos_prediction()

    # test test_intent() (non-bulk)
    test_ds = NluDataset(["Ich habe kein DSL und telefon"])
    watson.set_bulk(False)
    assert watson._use_bulk is False
    preds = watson.test_intent(test_ds)
    assert isinstance(preds, list)
    assert len(preds) == 1
    assert preds[0] in train_data.unique_intents

    # test n_best_intents argument to test_intent()
    preds, probs = watson.test_intent(test_ds, return_probs=True, n_best_intents=5)
    assert isinstance(preds, list)
    assert isinstance(probs, list)
    assert isinstance(preds[0], list)
    assert isinstance(probs[0], list)
    assert preds[0][0] in train_data.unique_intents
    assert len(preds[0]) == 2
    assert isinstance(probs[0][0], float) or (probs[0][0] == 1)
    assert len(probs[0]) == 2



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import logging
import random

import spacy
from spacy.pipeline.textcat import DEFAULT_SINGLE_TEXTCAT_MODEL
from spacy.training import Example

from .vendor import Vendor


logger = logging.getLogger(__name__)


class Spacy(Vendor):
    def __init__(self, n_iter=10, config=None, language="en"):
        """
        Interface for the Spacy intent classifier.

        :param n_iter: Number of training iterations (default: 1000)
        :type n_iter: int
        :param config: Textcat config in dictionary format as described
            at https://spacy.io/api/textcategorizer
        :type config: dict
        :param language: Language string (default: "en")
        :type language: str
        """
        self._alias = self.name
        self._nlp = spacy.blank(language)
        if not config:
            config = {
                "threshold": 0.5,
                "model": DEFAULT_SINGLE_TEXTCAT_MODEL,
            }
        self._textcat = self._nlp.add_pipe("textcat", config=config)
        self._n_iter = n_iter

    def _convert(self, dataset):
        examples = []
        for text, intent in zip(dataset.texts, dataset.intents):
            doc = self._nlp.make_doc(text)
            cats = {
                intent: True if intent == each else False
                for each in dataset.unique_intents
            }
            example = Example.from_dict(doc, {"cats": cats})
            examples.append(example)
        return examples

    def train_intent(self, dataset):
        """Train intent classifier."""
        examples = self._convert(dataset)

        get_examples = lambda: examples  # noqa: E731
        optimizer = self._nlp.initialize(get_examples)
        for itn in range(self._n_iter):
            random.shuffle(examples)
            for example in examples:
                self._nlp.update([example], sgd=optimizer)

    def test_intent(self, dataset, return_probs=False):
        """Test intent classifier."""
        intents = []
        probs = []
        for text in dataset.texts:
            doc = self._nlp(text)
            intent = max(doc.cats, key=doc.cats.get)
            prob = doc.cats[intent]
            intents.append(intent)
            probs.append(prob)
        if return_probs:
            return intents, probs
        return intents



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import json
import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime

import requests
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_watson import AssistantV1
from tqdm import tqdm

from nlubridge.nlu_dataset import OUT_OF_SCOPE_TOKEN

from .vendor import Vendor


logger = logging.getLogger(__name__)


class WatsonAssistant(Vendor):
    def __init__(
        self,
        api_key=None,
        endpoint=None,
        workspace_name=None,
        api_version="2018-02-16",
        max_workers=10,
        use_bulk=True,
    ):
        """
        Interface for IBM Watson Assistant.

        See `IBM Watson Assistant
        <https://www.ibm.com/watson/services/conversation/>`_.

        Uses either of two methods for predictions:
            * use_bulk=True -> uses bulk_classify method of the WA API.
              This is free of charge but available only in premium
              subscriptions. It is also rate limited, possibly making
              it slower in some cases where dataset is long.
              Rate limit parameters are hard-coded, i.e. we need to
              change them in the code if the API changes
            * use_bulk=False -> Makes concurrent requests to the
              message WA API method. For large numbers of test cases,
              this can become expensive.

        :param api_key: see Watson Assistant v1 API docs
        :type api_key: str
        :param endpoint:see Watson Assistant v1 API docs
        :type endpoint: str
        :param workspace_name: see Watson Assistant v1 API docs
        :type workspace_name: str
        :param api_version: see Watson Assistant v1 API docs
        :type api_version: str
        :param max_workers: number of ThreadPoolExecutor workers when use_bulk=False
            (default=10)
        :type max_workers: int
        :param use_bulk: if True (default) uses bulk_classify method
        :type use_bulk: bool
        """
        self._alias = self.name
        api_key = api_key or os.getenv("WATSON_API_KEY")
        if api_key is None:
            ValueError(
                "api_key not passed and not found under environment variable"
                "WATSON_API_KEY"
            )
        endpoint = endpoint or os.getenv("WATSON_ENDPOINT")
        if endpoint is None:
            ValueError(
                "endpoint not passed and not found under environment variable"
                "WATSON_ENDPOINT"
            )
        workspace_name = workspace_name or os.getenv("WATSON_WORKSPACE_NAME")
        if workspace_name is None:
            ValueError(
                "workspace_name not passed and not found under environment "
                "variable WATSON_WORKSPACE_NAME"
            )
        self._assistant = self._connect(endpoint, api_key, api_version)
        self._workspace_name = workspace_name
        self._workspace_id = self._get_workspace_id()
        self._max_workers = max_workers
        self._use_bulk = use_bulk

    def set_bulk(self, use_bulk):
        """
        Set use_bulk on (True) or off (False).

        :param use_bulk: True means use bulk method for testing
        :type use_bulk: bool
        """
        self._use_bulk = use_bulk

    def set_max_workers(self, max_workers):
        """
        Set max_workers for parallel API requests.

        Applicable for processing of test requests if use_bulk=False.

        :param max_workers: max_workers to use for parallel requests
        :type max_workers: int
        """
        self._max_workers = max_workers

    @property
    def _is_trained(self):
        response = self._assistant.get_workspace(
            workspace_id=self._workspace_id, export=False
        ).get_result()
        return response["status"] == "Available"

    def train_intent(self, dataset):
        """Train intent classifier."""
        logger.info(f"Training on {dataset.n_samples} samples")
        self._upload_samples(dataset)
        while not self._is_trained:
            time.sleep(3)
        # Return vendor for compatibility with other vendors implementations
        return self

    def test_intent(
        self,
        dataset,
        return_probs=False,
        max_workers=None,
        n_best_intents=1,
        use_bulk=None,
    ):
        """
        Test intent classifier.

        :param dataset: nlutests dataset
        :type dataset: nlutests.datasets.NLUdataset
        :param return_probs: if True, returns probability of predicted intent
            (default=False)
        :type return_probs: bool
        :param n_best_intents: number of n-best results to return (default=1)
        :type n_best_intents: int
        :param max_workers: deprecated; pass to constructor instead
        :param use_bulk: deprecated; pass to constructor instead
        :return: predictions, probabilities
        """
        if (use_bulk is not None) or (max_workers is not None):
            logger.warning(
                "Deprecated: Please do not pass use_bulk and "
                "max_workers to test_intent anymore but to the "
                "constructor instead. Settings will not apply "
                "if passed to test_intent!"
            )

        logger.info(f"Testing on {dataset.n_samples} samples")

        intents = []
        probs = []
        if not self._is_trained:
            raise RuntimeError("Watson Assistant workspace has not yet been trained")

        if not self._use_bulk:
            responses = self._get_wa_response_in_session(dataset)
        else:
            responses = self._get_wa_response_in_batches(dataset)

        for text, result in zip(dataset.texts, responses):
            intent, prob = self._extract_top_intent_from_result(result, n_best_intents)
            intents.append(intent)
            probs.append(prob)
        if return_probs:
            return intents, probs
        return intents

    @staticmethod
    def validate_text(text):
        """
        Validate an utterance text.

        Validate the test data text so it conforms to the
        Watson Assistant API format, see
        https://www.ibm.com/watson/developercloud/conversation/api/v1/?python#send_message

        :param text: User input
        :type text: str
        """
        val_text = text
        not_allowed = ["\n", "\t", "\r"]
        if len(val_text) > 2048:
            val_text = val_text[0:2047]
        if any(x in val_text for x in not_allowed):
            for char in not_allowed:
                val_text = val_text.replace(char, " ")
        return val_text

    def _connect(self, endpoint, api_key, api_version):
        authenticator = IAMAuthenticator(api_key)
        assistant = AssistantV1(version=api_version, authenticator=authenticator)
        assistant.set_service_url(endpoint)
        return assistant

    def _convert(self, dataset):
        """
        Convert data from the standardized format to Watson Assistant format.

        This function accepts a dataset as input and returns a list of
        samples in the vendor specific expected format.
        https://www.ibm.com/watson/developercloud/conversation/api/v1/?python#create_example

        Texts are modified with self.validate_text to conform to WA
        requirements. In case of duplicates (case-insensitive), only
        the first encountered sample is kept.
        """
        intents_examples = {}
        exist_texts = []
        for text, intent in zip(dataset.texts, dataset.intents):
            text = self.validate_text(text)
            if text.lower() in exist_texts:
                # it's a duplicate text
                continue
            else:
                exist_texts.append(text.lower())
                try:
                    intents_examples[intent].append({"text": text})
                except KeyError:
                    intents_examples[intent] = [{"text": text}]

        data_wa_format = [
            {"intent": intent, "examples": examples}
            for (intent, examples) in intents_examples.items()
        ]
        return data_wa_format

    def _get_workspace_id(self):
        response = self._assistant.list_workspaces().get_result()
        for workspace in response["workspaces"]:
            if workspace["name"] == self._workspace_name:
                return workspace["workspace_id"]
        raise ValueError("Workspace could not be found.")

    def _delete_workspace(self):
        """Delete a workspace by its name."""
        response = self._assistant.delete_workspace(self._workspace_id)
        status_code = response.get_status_code()
        return status_code == "200"

    def _clear_workspace(self):
        """
        Create a fresh WA workspace with the same name as used before.

        The intents in the workspace need to be deleted before a new
        test uploads new intents. This method deletes the existing
        workspace and creates a new one under the same name
        (workspace_id is updated).
        """
        logger.info(f"Clearing workspace {self._workspace_id}")
        # TODO: handle failures of delete_workspace (and fix unused variable 'success')
        success = self._delete_workspace()  # noqa: F841
        response = self._assistant.create_workspace(
            name=self._workspace_name, description="created via api", language="de"
        ).get_result()
        self._workspace_id = response["workspace_id"]

    def _upload_samples(self, dataset):
        self._clear_workspace()
        intents = self._convert(dataset)
        logger.info("Uploading samples")
        for intent in intents:
            # TODO: handle failures (and fix unused variable 'response')
            response = self._assistant.create_intent(  # noqa: F841
                workspace_id=self._workspace_id,
                intent=intent["intent"],
                examples=intent["examples"],
            ).get_result()

    def _get_wa_response(self, query):
        """Get the NLU result for a single message."""
        query = self.validate_text(query)
        response = self._assistant.message(
            workspace_id=self._workspace_id,
            input={"text": query},
            alternate_intents=True,
        )
        return response.get_result()

    def _get_wa_response_in_batches(self, dataset):
        """Get intent predictions using the bulk_classify method."""
        logger.info("Using /bulk_classify endpoint for testing")

        def get_response(queries):
            data = {"input": [{"text": self.validate_text(query)} for query in queries]}
            status_code = 429
            while status_code == 429:
                response = session.post(url, data=json.dumps(data))
                status_code = response.status_code
                if status_code == 429:
                    reset_time = datetime.fromtimestamp(
                        int(response.headers.get("X-RateLimit-Reset", None))
                    )
                    now = datetime.now()
                    wait_seconds = (reset_time - now).total_seconds()
                    logger.info(f"waiting {wait_seconds} seconds (rate limit was hit)")
                    time.sleep(wait_seconds)
            return response

        with requests.Session() as session:
            session.auth = requests.auth.HTTPBasicAuth(
                "apikey", self._assistant.authenticator.token_manager.apikey
            )
            session.headers.update({"Content-Type": "application/json"})
            url = "{url}/v1/workspaces/{ws_id}/bulk_classify?version={version}".format(
                url=self._assistant.service_url,
                ws_id=self._workspace_id,
                version=self._assistant.version,
            )
            all_results = []
            for i in range(0, len(dataset.texts), 50):
                batch = dataset.texts[i : i + 50]
                response = get_response(batch)
                results = response.json()["output"]
                # ensure that dataset.texts and results are correctly aligned
                assert batch == [r["input"]["text"] for r in results]
                all_results += results
        return all_results

    def _get_wa_response_in_session(self, dataset):
        """Get intent predictions with parallelized request."""
        logger.info("Using /message endpoint for testing. Costs may apply!")

        def get_response(query):
            data = {"input": {"text": query}, "alternate_intents": True}
            response = session.post(url, data=json.dumps(data))
            return response

        with requests.Session() as session:
            session.auth = requests.auth.HTTPBasicAuth(
                "apikey", self._assistant.authenticator.token_manager.apikey
            )
            session.headers.update({"Content-Type": "application/json"})
            url = "{url}/v1/workspaces/{ws_id}/message?version={version}".format(
                url=self._assistant.service_url,
                ws_id=self._workspace_id,
                version=self._assistant.version,
            )
            with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
                responses = list(
                    tqdm(
                        executor.map(get_response, dataset.texts),
                        total=len(dataset.texts),
                    )
                )
            all_responses = []
            for response in responses:
                try:
                    r = response.json()
                except requests.exceptions.JSONDecodeError:
                    r = {"error": response.text}
                all_responses.append(r)
            return all_responses

    @staticmethod
    def _extract_top_intent_from_result(wa_result, n_best_intents):
        result_intents = wa_result.get("intents")
        if result_intents:
            if n_best_intents == 1:
                intent = result_intents[0].get("intent")
                prob = result_intents[0].get("confidence")
            else:
                intent = []
                prob = []
                n_best_max = (
                    len(result_intents)
                    if n_best_intents >= len(result_intents)
                    else n_best_intents
                )
                for k in range(0, n_best_max):
                    intent.append(result_intents[k].get("intent"))
                    prob.append(result_intents[k].get("confidence"))
        elif wa_result.get("error"):
            intent = wa_result.get("error")
            prob = -1
        else:
            intent = OUT_OF_SCOPE_TOKEN
            prob = 0
        return intent, prob



# Copyright (c) 2022 Ralf Kirchherr, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT
from __future__ import annotations

import logging
import os
import pathlib
import tempfile
from typing import List, Optional, Tuple, Union

from rasa.core.agent import Agent
from rasa.core.channels.channel import UserMessage
from rasa.model import get_local_model
from rasa.model_training import train_nlu
from rasa.shared.nlu.constants import (
    INTENT,
    INTENT_NAME_KEY,
    INTENT_RANKING_KEY,
    PREDICTED_CONFIDENCE_KEY,
)

from nlubridge import NBestKeys, NluDataset, to_rasa
from nlubridge.dataloaders.rasa import convert_entities_to_nludataset
from nlubridge.vendors.vendor import Vendor


logger = logging.getLogger(__name__)

DEFAULT_INTENT_RASA_CONFIG_PATH = os.path.join(
    pathlib.Path(__file__).parent.absolute(), "config", "rasa_nlu_config.yml"
)


class Rasa3(Vendor):
    def __init__(self, model_config: Optional[str] = None):
        """
        Interface for the `Rasa NLU <https://github.com/RasaHQ/rasa>`_.

        Uses the default pipeline as of January 22 2021. No algorithmic
        details have been researched. The configuration file can be
        found in the config directory within this directory. A custom
        pipeline can be provided as argument to the class constructor.

        :param model_config: filepath to a Rasa config file
        """
        self._alias = self.name
        self._config = model_config
        self._agent = None

    @classmethod
    def from_model(cls, model_path: Union[str, pathlib.Path]) -> Rasa3:
        """
        Construct a Rasa3 object from a saved model.

        Loading a model can fail if versions of dependencies differ between the
        environment where the model was trained and the environment where the model
        is loaded.

        :param model_path: file path to a saved rasa model.
        :return: Rasa
        """
        obj = cls()
        # Since we don't know the config used to train the model, make sure we don't
        # add an inconsistent config accidentally should we ever change __init__().
        assert obj._config is None
        model_archive = get_local_model(model_path)
        obj._agent = Agent.load(model_path=model_archive)
        return obj

    def train(self, dataset: NluDataset) -> Rasa3:
        """
        Train intent and/or entity classification.

        :param dataset: Training data
        :return: It's own Rasa3 object
        """
        self._config = self._config if self._config else DEFAULT_INTENT_RASA_CONFIG_PATH
        with tempfile.TemporaryDirectory() as tmpdirname:
            nlu_yml_file = os.path.join(
                pathlib.Path(tmpdirname), "nlu.yml"
            )  # output path for temporary nlu.yml
            to_rasa(dataset, nlu_yml_file)
            logger.info(f"Start training (using {self._config!r})...")
            model_archive = train_nlu(self._config, nlu_yml_file, tmpdirname)
            logger.info("Training completed!")

            logger.info("Load model...")
            self._agent = Agent.load(model_path=model_archive)
            logger.info("Model loaded!")
        return self

    def train_intent(self, dataset: NluDataset) -> Rasa3:
        """
        Train intent classification.

        This method is mainly for compatibility reasons, as it in case of Rasa identical
        to the `train` method.

        :param dataset: Training data
        :return: It's own Rasa3 object
        """
        return self.train(dataset)

    def _parse_message(self, text):
        # result = asyncio.run(
        #     self._agent.parse_message(text)
        # )  # agent's parse method is a coroutine

        # To avoid errors in Jupyter if  an event loop is already running, don't use the
        # async method from self._agent
        message = UserMessage(text)
        parse_data = self._agent.processor._parse_message_with_graph(message)
        self._agent.processor._check_for_unseen_features(parse_data)
        result = parse_data
        return result

    def test(self, dataset: NluDataset) -> NluDataset:
        """
        Test a given dataset.

        Test a given dataset and obtain the intent and/or entity classification results
        in the NLUdataset format

        :param dataset: Input dataset to be tested
        :return: NLUdataset object comprising the classification results. The list of
            the predicted intent classification probabilities are accessible via the
            additional attribute 'probs' (List[float]).
        """
        if self._agent is None:
            raise RuntimeError("Rasa3 classifier has to be trained first!")
        intents = []
        n_best_lists = []
        entities_list = []
        for text in dataset.texts:
            result = self._parse_message(text)
            intent = result.get(INTENT, {}).get(INTENT_NAME_KEY)
            entities = convert_entities_to_nludataset(result)
            nbest = [
                {
                    NBestKeys.INTENT: ranked.get(INTENT_NAME_KEY),
                    NBestKeys.CONFIDENCE: ranked.get(PREDICTED_CONFIDENCE_KEY),
                }
                for ranked in result.get(INTENT_RANKING_KEY, [])
            ]

            intents.append(intent)
            n_best_lists.append(nbest)
            entities_list.append(entities)

        res = NluDataset(dataset.texts, intents, entities_list, n_best_lists)
        return res

    def test_intent(
        self, dataset: NluDataset, return_probs: bool = False
    ) -> Union[List[str], Tuple[List[str], List[float]]]:
        """
        Test a given dataset and obtain just the intent classification results.

        :param dataset: The dataset to be tested
        :param return_probs: Specifies if the probability values should be returned
            (default is False)
        :return: Either a list of predicted intent classification results or a tuple of
            predicted intent classification and probabilites results (depeding on
            argument 'return_probs')
        """
        if self._agent is None:
            raise RuntimeError("Rasa3 classifier has to be trained first!")
        intents = []
        probs = []
        for text in dataset.texts:
            result = self._parse_message(text)
            intent = result.get(INTENT, {}).get(INTENT_NAME_KEY)
            prob = result.get(INTENT, {}).get(PREDICTED_CONFIDENCE_KEY)
            intents.append(intent)
            probs.append(prob)
        if return_probs:
            res = (intents, probs)
        else:
            res = intents
        return res



import json

import requests
from test_vendors import (
    assert_multiple_utterances_predicted,
    assert_preds_are_intents,
    assert_return_probs,
)

from nlubridge import NluDataset
from nlubridge.vendors.watson_assistant import WatsonAssistant


FAKE_URL = "dummy.org"
FAKE_KEY = "dummy_key"
WS_NAME = "dummy_name"
WS_ID = 1


def test_watson_mocked(train_data, mocker):
    # TODO: Also mock tests for use_bulk=False (see test_watson_vendor.py)

    # Mock the ibm_watson AssistantV1 API for requests implemented by this
    mocker.patch("nlubridge.vendors.watson_assistant.AssistantV1", AssitantV1Mock)

    # Mock requests.Session.post() requests
    mocker.patch("requests.Session", SessionMock)

    # test initialization
    watson = WatsonAssistant(FAKE_URL, FAKE_KEY, WS_NAME)

    # test train_intent()
    watson.train_intent(train_data)

    # test test_intent() and properties of its return values
    assert_preds_are_intents(watson, train_data.unique_intents)
    assert_return_probs(watson, train_data.unique_intents)
    assert_multiple_utterances_predicted(watson, train_data)

    # Test for datasets that need to be passed in multiple batches
    # TODO: test handling of rate-limit (only 250/min allowed)
    cases90 = train_data + train_data + train_data
    cases360 = cases90 + cases90 + cases90 + cases90
    preds, probs = watson.test_intent(cases360, return_probs=True)
    assert len(preds) == 360
    assert len(probs) == 360

    test_ds = NluDataset(["Ich habe kein DSL und telefon"])

    # test n_best_intents argument to test_intent()
    preds, probs = watson.test_intent(test_ds, return_probs=True, n_best_intents=5)
    assert isinstance(preds, list)
    assert isinstance(probs, list)
    assert isinstance(preds[0], list)
    assert isinstance(probs[0], list)
    assert preds[0][0] in train_data.unique_intents
    assert len(preds[0]) == 2
    assert isinstance(probs[0][0], float) or (probs[0][0] == 1)
    assert len(probs[0]) == 2

    # test watson.set_bulk()
    watson.set_bulk(False)
    assert watson._use_bulk is False


class ResultMock:
    """
    We sometimes need to be able to call get_result() on a return value from
    AssistantV1Mock.
    """

    def __init__(self, result, status_code=None):
        self.result = result
        self.status_code = status_code

    def get_result(self):
        return self.result

    def get_status_code(self):
        return self.result

    def json(self):
        return self.result


class TokenManager:
    apikey = FAKE_KEY


class Authenticator:
    token_manager = TokenManager()


class AssitantV1Mock:
    """
    Mocks AssistantV1 from the ibm_watson package
    """

    service_url = FAKE_URL
    version = "111"
    authenticator = Authenticator()
    workspace_name = WS_NAME

    def __init__(self, **kwargs):
        return None

    @staticmethod
    def set_service_url(endpoint):
        pass

    @staticmethod
    def get_workspace(workspace_id, export):
        return ResultMock({"status": "Available"})

    @staticmethod
    def list_workspaces():
        result = {
            "workspaces": [
                {"name": WS_NAME, "workspace_id": WS_ID},
                {"name": "workspace2", "workspace_id": "2"},
            ]
        }
        return ResultMock(result)

    @staticmethod
    def delete_workspace(id):
        return ResultMock(200)

    @staticmethod
    def create_workspace(name, description, language):
        return ResultMock({"workspace_id": WS_ID})

    @staticmethod
    def create_intent(workspace_id, intent, examples):
        return ResultMock(None)


class SessionMock(requests.Session):
    """
    Mocks requests.Session.post()
    """

    def post(self, url, data):
        data_dict = json.loads(data)
        texts = [item["text"] for item in data_dict["input"]]
        return ResultMock(
            {
                "output": [
                    {
                        "input": {"text": text},
                        "entities": [],
                        "intents": [
                            {"intent": "help", "confidence": 0.061},
                            {"intent": "affirm", "confidence": 0.0448},
                        ],
                    }
                    for text in texts
                ]
            },
            200,
        )



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import collections
import csv
import itertools
import json
import numbers
import random
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
from sklearn.model_selection import StratifiedKFold, train_test_split


DATA_PATH = Path(__file__).parents[2] / "data"
OUT_OF_SCOPE_TOKEN = "out_of_scope"


class EntityKeys:
    """Keys for entity dicts in NLUdataset."""

    TYPE = "entity"
    START = "start"
    END = "end"


class NLUdataset:
    def __init__(
        self,
        texts: List[str],
        intents: Optional[List[str]] = None,
        entities: Optional[List[List[dict]]] = None,
        out_of_scope=False,
        max_intent_length=None,
        seed=42,
    ):
        """
        Class for managing NLU data with intent and entity annotations.

        :param texts: A list of utterances
        :type texts: List[str]
        :param intents: A list of intents for each text. If intents is not
            empty then there needs to be an intent for each text.
        :type intents: List[str]
        :param entities: If intents is not empty then there needs to be an
            intent for each text.
        :type entities: List[dict]
        :param max_intent_length: If given the intents will be strimmed to
            the first `max_intent_length` in chars. This is required
            because some vendors like LUIS don't accept very long intent names.
        :type max_intent_length: int or None
        :param seed: The initial random seed used by shuffle.
        :type seed: int
        """
        texts = list(texts) if texts is not None else []
        intents = list(intents) if intents is not None else None
        entities = list(entities) if entities is not None else None

        self.texts = texts
        self.n_samples = len(self.texts)
        self.intents: List[Optional[str]]
        self.entities: List[List[Dict]]

        if not intents or intents is None:
            self.unique_intents = []
            if out_of_scope:
                self.unique_intents = [OUT_OF_SCOPE_TOKEN]
            self.n_intents = len(self.unique_intents)
            self.intents = [None for _ in texts]
        else:
            if max_intent_length:
                cropped_intents = [intent[:max_intent_length] for intent in intents]
                if len(set(intents)) > len(set(cropped_intents)):
                    warnings.warn(
                        "Intent names are not unique on first {} characters".format(
                            max_intent_length
                        )
                    )
                intents = cropped_intents
            # Make copy so typing is consistent (makes sure we don't accidentally add
            # None elements to the original list from within this class)
            self.intents = list(intents)
            self.unique_intents = list(dict.fromkeys(intents))

            if out_of_scope:
                self.unique_intents += [OUT_OF_SCOPE_TOKEN]

            self.n_intents = len(self.unique_intents)
            self.intent_frequencies = collections.Counter(self.intents)

        if not entities:
            self.entities = [list() for _ in texts]
            self.unique_entities = []
            self.n_entities = 0
        else:
            self.entities = entities

        if entities is not None:
            it = itertools.chain.from_iterable(entities)
            unique_entities = []
            for each in it:
                if isinstance(each, dict) and EntityKeys.TYPE in each:
                    unique_entities.append(each[EntityKeys.TYPE])
            self.unique_entities = list(dict.fromkeys(unique_entities))
            self.n_entities = len(self.unique_entities)

        self.data = [
            (text, intent, entities)
            for text, intent, entities in zip(self.texts, self.intents, self.entities)
        ]
        # TODO: Can we handle random seed outside this class to avoid confusions
        #  when making implicit copies like in sample() method?
        random.seed(seed)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, key):
        if isinstance(key, slice):
            texts = self.texts[key]
            intents = self.intents[key]
            entities = self.entities[key]
            return NLUdataset(texts, intents, entities)

        elif isinstance(key, collections.abc.Sequence) or isinstance(key, np.ndarray):
            texts = [self.texts[each] for each in key]
            intents = [self.intents[each] for each in key]
            entities = [self.entities[each] for each in key]
            return NLUdataset(texts, intents, entities)

        elif isinstance(key, numbers.Integral):
            return self.data[key]

    def __len__(self):
        return len(self.texts)

    def __add__(self, other):
        return NLUdataset(
            texts=self.texts + other.texts,
            intents=self.intents + other.intents,
            entities=self.entities + other.entities,
        )

    @classmethod
    def from_joined(cls, *datasets):
        """Join NLUdatasets provided in a list."""
        texts, intents, entities = [], [], []

        for dataset in datasets:
            texts.extend(dataset.texts)
            intents.extend(dataset.intents)
            entities.extend(dataset.entities)

        joint_dataset = cls(texts, intents, entities)
        return joint_dataset

    @property
    def name(self):
        """Return the class name."""
        return self.__class__.__name__

    def shuffle(self):
        """Shuffles the dataset."""
        random.shuffle(self.data)
        self.texts, self.intents, self.entities = zip(*self.data)
        return self

    def sample(self, size=0.1, random_state=0, stratification="intents"):
        """
        Sample the dataset preserving intent proportions.

        :param size: can be an integer representing the final number of
            samples or a float representing the ratio
        :type size: float or int
        :param random_state: seed for random processes
        :type random_state: int
        :param stratification: If not None, data is split in a stratified fashion, using
            this as the class labels. Default is using the intent labels for
            stratification.
        """
        if isinstance(size, numbers.Integral):
            size = size / self.n_samples

        _, sampled = self.train_test_split(
            test_size=size, random_state=random_state, stratification=stratification
        )
        return sampled

    def filter_by_intent_name(
        self,
        excluded: Optional[List[str]] = None,
        allowed: Optional[List[str]] = None,
    ):
        """
        Filter the dataset by intents.

        :param excluded: List of intents for which not to return data
        :type excluded: List[str]
        :param allowed: List of intents for which to return data
        :type allowed: List[str]
        """
        if excluded is None:
            excluded = []

        if allowed is None:
            allowed = self.unique_intents

        texts, intents, entities = [], [], []
        for text, intent, entity_list in self.data:
            if intent in excluded:
                continue

            if intent not in allowed:
                continue

            texts.append(text)
            intents.append(intent)
            entities.append(entity_list)

        return NLUdataset(texts, intents, None)

    def select_first_n_intents(self, n: int):
        """
        Return a dataset with a max of `n` intents from the dataset.

        The order of the selected intents is given by
        `self.unique_intents`

        :param n: Number of intents to return
        :type n: int
        """
        allowed = self.unique_intents[:n]
        ds = self.filter_by_intent_name(allowed=allowed)
        return ds

    def clip_by_intent_frequency(self, max_frequency, min_frequency=None):
        """
        Sample the dataset leaving only `max_freq` samples per intent.

        If min_frequency is given then intents with less than that
        number of samples will be dropped.

        Example:
        -------
        To get a dataset with all intents having the same number of
        samples you can do:

        >>> from nlutests.datasets import TDGDataset
        >>> ds = TDGDataset()
        >>> ds.intent_frequencies
        Counter({'telefonie_disambiguierung': 562,
                'telefonie_beidseitig': 869,
                'internet_kein_internet': 815,
                'internet_langsam': 764,
                'internet_abbrueche': 654,
                'komplettausfall_eindeutig': 931})

        >>> ds.clip_by_intent_frequency(700, 700).intent_frequencies
        Counter({'telefonie_beidseitig': 700,
                'internet_kein_internet': 700,
                'internet_langsam': 700,
                'komplettausfall_eindeutig': 700})

        :param max_frequency: maximum number of samples per intent to
            include in the returned dataset
        :type max_frequency: int
        :param min_frequency: maximum number of samples per intent to
            include in the returned dataset. If an intent has less
            samples, it will be dropped.
        :type min_frequency: int

        """
        freqs = collections.Counter()
        texts, intents = [], []

        for text, intent, _ in self.data:
            if (
                min_frequency is not None
                and self.intent_frequencies[intent] <= min_frequency
            ):
                continue

            if freqs[intent] == max_frequency:
                continue

            freqs[intent] += 1
            texts.append(text)
            intents.append(intent)

        return NLUdataset(texts, intents, None)

    def subsample_by_intent_frequency(self, target_rate, min_frequency, shuffle=False):
        """
        Return a smaller dataset with similar intent distribution.

        Only a certain percentage of each intent's data is kept, while
        for intents with fewer examples that percentage is higher.

        Intents with less than min_frequency samples will not be
        reduced. For all others, the target count is computed as
        `min_freq + (num_samples - min_freq) * target_rate`
        """
        if shuffle:
            raise NotImplementedError(
                "shuffle=True has not yet been implemented. You can "
                "use NLUdataset.shuffle() to shuffle the dataset "
                "before subsampling."
            )
        freqs = collections.Counter()
        texts, intents, entities = [], [], []
        for text, intent, entity_list in self.data:
            intent_freq = self.intent_frequencies[intent]
            target_freq = min_frequency + ((intent_freq - min_frequency) * target_rate)
            is_below_min = intent_freq <= min_frequency
            is_below_target = freqs[intent] < target_freq
            if is_below_min or is_below_target:
                freqs[intent] += 1
                texts.append(text)
                intents.append(intent)
                entities.append(entity_list)
        return NLUdataset(texts, intents, entities)

    def train_test_split(
        self, test_size=0.25, random_state=0, stratification="intents", **args
    ):
        """
        Split dataset into train and test partitions.

        :param test_size: fraction of samples to use for testing
        :type test_size: float
        :param random_state: random seed
        :type random_state: int
        :param stratification: TODO
        :type stratification: TODO
        :param args: additional args for sklearn's train_test_split
            provided as dictionary
        :type args: dict
        """

        def configure_stratification():
            """
            Allow configuring stratification options.

            The default setting, "intents", is used when we want to
            use stratification by intent strings. If we want to use
            anything else, we can specify it in the (parent) method's
            stratification parameter, e.g. None, which would turn
            stratification off (using the underlying
            sklearn.model_selection.train_test_split() method).

            The default setting, which is passed as a string, will be
            converted into the appropriate self.intents, which isn't
            accessible from the parent methodd's parameters.

            Returns the stratify setting (either self.intents or
            whatever is passed) to the sklearn train_test_split
            method.
            """
            if stratification == "intents":
                stratify = self.intents
            else:
                stratify = stratification
            return stratify

        (
            texts_train,
            texts_test,
            intents_train,
            intents_test,
            entities_train,
            entities_test,
        ) = train_test_split(
            self.texts,
            self.intents,
            self.entities,
            stratify=configure_stratification(),
            test_size=test_size,
            random_state=random_state,
            **args,
        )

        train_ds = NLUdataset(texts_train, intents_train, entities_train)
        test_ds = NLUdataset(texts_test, intents_test, entities_test)
        return train_ds, test_ds

    def _filter_by_index_list(self, idx_list):
        filtered = [d for (idx, d) in enumerate(self.data) if idx in idx_list]
        texts, intents, entities = zip(*filtered)
        return NLUdataset(texts, intents, entities)

    def cross_validation_splits(self, cv_iterator=None):
        """Cross validation generator function."""
        if cv_iterator is None:
            cv_iterator = StratifiedKFold(n_splits=5, shuffle=True)
        for train_idx, test_idx in cv_iterator.split(self.texts, self.intents):
            train_ds = self._filter_by_index_list(train_idx)
            test_ds = self._filter_by_index_list(test_idx)
            yield train_ds, test_ds

    def to_records(self):
        """
        Return data as a list of dicts.

        Each dict represents a single utterance/document.
        """
        examples = []
        for text, intent, entities in self.data:
            example = {
                "text": text,
                "intent": intent,
                "entities": [
                    {
                        "entity": entity[EntityKeys.TYPE],
                        "start": entity[EntityKeys.START],
                        "end": entity[EntityKeys.END],
                    }
                    for entity in entities
                ],
            }  # yapf: disable
            examples.append(example)
        return examples

    def to_dict(self):
        """
        Return intent data as dict.

        Intents are keys, each holding a list of corresponding
        utterances.
        """
        d = collections.defaultdict(list)
        for text, intent, _ in self.data:
            d[intent].append(text)

        return dict(d)

    def to_json(self, path=None, records=None):
        """
        Convert dataset to JSON.

        If a path is given, saves to that file. Otherwise returns a
        JSONstring. Records can be passed explicitly, otherwise
        self.to_records() will be called to obtain the data in
        dictionary format.

        :param path: optional path under which to save the JSON file
        :type path: str
        :param records: optional JSON-serializable data
        :type records: dict or list
        """
        if records is None:
            records = self.to_records()

        if path is None:
            return json.dumps(records, indent=4, ensure_ascii=False)

        with open(path, "w") as f:
            json.dump(records, f, indent=4, ensure_ascii=False)


def from_csv(filepath, text_col, intent_col) -> NLUdataset:
    """Load dataset (only text and intents) from a csv file."""
    columns: Dict[Union[int, str], List] = collections.defaultdict(list)

    with open(filepath) as f:
        reader = csv.DictReader(f)
        for row in reader:
            for k, v in row.items():
                columns[k].append(v)

    if isinstance(text_col, int):
        key = list(columns.keys())[text_col]
        columns[text_col] = [key, *columns[key]]

    if isinstance(intent_col, int):
        key = list(columns.keys())[intent_col]
        columns[intent_col] = [key, *columns[key]]

    ds = NLUdataset(columns[text_col], columns[intent_col])
    return ds


def from_json(
    json_string,
    text_key="text",
    intent_key="intent",
    entities_key="entities",
    entity_type_key="entity",
    entity_start_key="start",
    entity_end_key="end",
    end_index_add_1=False,
):
    """
    Load the dataset form a json string.

    The json string should have the structure in the TestDataset class.
    """

    def format_entities(entities, type_key, start_key, end_key, end_index_add_1):
        ex_entities = []
        for entity in entities:
            formatted_entity = {
                EntityKeys.TYPE: entity[type_key],
                EntityKeys.START: entity[start_key],
                EntityKeys.END: entity[end_key] + end_index_add_1,
            }
            # Add any custom keys defined in the source json
            for key in entity.keys():
                if key not in [type_key, start_key, end_key]:
                    formatted_entity[key] = entity[key]
            ex_entities.append(formatted_entity)
        return ex_entities

    examples = json.loads(json_string)
    texts, intents, entities = [], [], []
    for example in examples:
        texts.append(example[text_key])
        intents.append(example[intent_key])
        example_entities = format_entities(
            example[entities_key],
            entity_type_key,
            entity_start_key,
            entity_end_key,
            end_index_add_1,
        )
        entities.append(example_entities)
    dataset = NLUdataset(texts, intents, entities)
    return dataset



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

from nlubridge.nlu_dataset import NluDataset

from .utils import from_csv


def from_watson_assistant(filepath) -> NluDataset:
    """Load data from Watson format as NLUdataset."""
    return from_csv(filepath, text_col=0, intent_col=1)



# Copyright (c) 2021 Klaus-Peter Engelbrecht, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Build script for setuptools."""

import os

from setuptools import find_packages, setup


project_name = "nlubridge"
source_code = "https://github.com/telekom/nlu-bridge"
keywords = [
    "nlu",
    "intent recognition",
    "natural language understanding",
    "evaluation",
    "performance",
]


def get_version():
    """Read version from ``__init__.py``."""
    version_filepath = os.path.join(
        os.path.dirname(__file__), project_name, "__init__.py"
    )
    with open(version_filepath) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split()[-1][1:-1]
    assert False


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name=project_name,
    version=get_version(),
    maintainer="Klaus-Peter Engelbrecht",
    author="Klaus-Peter Engelbrecht",
    author_email="k.engelbrecht@telekom.de",
    description=(
        "Provides a unified API to several popular intent recognition applications"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=keywords,
    url=source_code,
    project_urls={
        "Bug Tracker": source_code + "/issues",
        # "Documentation": "https://telekom.github.io/HPOflow/",
        "Source Code": source_code,
        "Contributing": source_code + "/blob/main/CONTRIBUTING.md",
        "Code of Conduct": source_code + "/blob/main/CODE_OF_CONDUCT.md",
    },
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=["scikit-learn", "python-dotenv", "lazy-imports", "ratelimit"],
    extras_require={
        "watson": ["ibm_watson", "tqdm", "requests"],
        "fasttext": ["fasttext"],
        "luis": ["requests", "ratelimit"],
        "rasa2": ["rasa~=2.8.0", "websockets~=10.4"],
        "rasa3": ["rasa==3.4.4", "websockets~=10.4"],
        "spacy": ["spacy==3.1.3"],
        "huggingface": ["datasets~=1.18"],
        "develop": [
            "pytest-cov",
            "pytest-mock",
            "flake8",
            "black",
            "pydocstyle",
            "setuptools",
            "wheel",
            "twine",
            "isort",
            "mdformat",
            "mypy",
            "pylint==2.14.5",
        ],
    },
    include_package_data=True,
)



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

from __future__ import annotations

import collections
import itertools
import json
import numbers
import random
import warnings
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import numpy as np
from sklearn.model_selection import (
    BaseCrossValidator,
    StratifiedKFold,
    train_test_split,
)


OUT_OF_SCOPE_TOKEN = "out_of_scope"


class EntityKeys:
    """Keys for entity dicts in NLUdataset."""

    TYPE = "entity"
    START = "start"
    END = "end"
    CONFIDENCE = "confidence"


class NBestKeys:
    """Keys for n-best list items (dicts) in NLUdataset."""

    INTENT = "intent"
    CONFIDENCE = "confidence"


class NluDataset:
    def __init__(
        self,
        texts: List[str],
        intents: Optional[List[str]] = None,
        entities: Optional[List[List[Dict]]] = None,
        n_best_intents: Optional[List[List[Dict]]] = None,
        max_intent_length=None,
    ) -> None:
        """
        Class for managing NLU data with intent and entity annotations.

        :param texts: A list of utterances
        :param intents: A list of intents, one for each text. If intents is
            not None then there needs to be an intent for each text.
        :param entities: List of entities for each text. If not None then
            there needs to be a List for each text.
        :param n_best_intents: List of n-best results (each denoted as a dict
            having intent and confidence keys) for each text. If not None then
            there needs to be a List for each text.
        :param max_intent_length: If given the intents will be trimmed to
            the first `max_intent_length` chars. This is required
            because some vendors like LUIS don't accept very long intent names.
        """
        # Cast to list so we can be a bit more flexible with the inputs, e.g. use a
        # Pandas Series
        texts = list(texts) if texts is not None else []
        intents = list(intents) if intents is not None else None
        entities = list(entities) if entities is not None else None

        ds_intents = (
            self._prepare_intents(max_intent_length, intents)
            if intents
            else [None for _ in texts]
        )
        ds_entities = entities if entities else [list() for _ in texts]
        ds_n_best_intents = (
            n_best_intents if n_best_intents else [list() for _ in texts]
        )
        self._data = list(zip(texts, ds_intents, ds_entities, ds_n_best_intents))
        self._unique_intents = self._get_unique_intents()
        self._intent_frequencies: collections.Counter = collections.Counter(
            self.intents
        )
        self._unique_entities = self._get_unique_entities()

    @property
    def name(self) -> str:
        """Return the class name."""
        return self.__class__.__name__

    @property
    def unique_intents(self) -> List[str]:  # noqa: D102
        return self._unique_intents

    @property
    def intent_frequencies(self) -> collections.Counter:  # noqa: D102
        return self._intent_frequencies

    @property
    def unique_entities(self) -> List[str]:  # noqa: D102
        return self._unique_entities

    @property
    def texts(self) -> List[str]:  # noqa: D102
        return [record[0] for record in self._data]

    @property
    def intents(self) -> List[str]:  # noqa: D102
        return [record[1] for record in self._data]

    @property
    def entities(self) -> List[List[Dict]]:  # noqa: D102
        return [record[2] for record in self._data]

    @property
    def confidences(self) -> List[Optional[float]]:  # noqa: D102
        try:
            confs = [record[3][0].get(NBestKeys.CONFIDENCE) for record in self._data]
        except IndexError:
            confs = [None for _ in self._data]
        return confs

    @property
    def n_best_intents(self) -> List[List[Dict]]:  # noqa: D102
        return [record[3] for record in self._data]

    @property
    def n_samples(self) -> int:  # noqa: D102
        return len(self._data)

    @property
    def n_intents(self) -> int:  # noqa: D102
        return len(self.unique_intents)

    @property
    def n_entities(self) -> int:  # noqa: D102
        return len(self.unique_entities)

    @classmethod
    def from_joined(cls, *datasets: NluDataset) -> NluDataset:
        """Join NLUdatasets provided in a list."""
        warnings.warn(
            "This method will be removed in a future version. Please use "
            "nlubridge.concat() or NluDataset.join().",
            DeprecationWarning,
        )
        data = []
        for dataset in datasets:
            data.extend(dataset._data)
        joint_dataset = NluDataset._from_data(data)
        return joint_dataset

    @classmethod
    def _from_data(cls, data):
        return cls(*zip(*data))

    def _get_unique_intents(self):
        # unique_intents = list(dict.fromkeys(intents))  # old
        unique_intents = sorted(
            [intent for intent in list(set(self.intents)) if intent is not None]
        )
        return unique_intents

    def _get_unique_entities(self):
        it = itertools.chain.from_iterable(self.entities)
        entity_types = [
            entity[EntityKeys.TYPE]
            for entity in it
            if isinstance(entity, dict) and EntityKeys.TYPE in entity
        ]
        return sorted(list(set(entity_types)))

    @staticmethod
    def _prepare_intents(max_intent_length, intents):
        if not max_intent_length:
            # Return a copy of intents to avoid interference with usages outside this
            # class; also makes sure we have a list
            return list(intents)
        cropped_intents = [intent[:max_intent_length] for intent in intents]
        if len(set(intents)) > len(set(cropped_intents)):
            warnings.warn(
                "Intent names are not unique on first {} characters".format(
                    max_intent_length
                )
            )
        return cropped_intents

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, key):
        if isinstance(key, slice):
            data = self._data[key]
            return NluDataset._from_data(data)

        elif isinstance(key, collections.abc.Sequence) or isinstance(key, np.ndarray):
            data = [self._data[each] for each in key]
            return NluDataset._from_data(data)

        elif isinstance(key, numbers.Integral):
            data = [self._data[key]]
            return NluDataset._from_data(data)

    def __len__(self):
        return len(self._data)

    def __add__(self, other):
        return NluDataset._from_data(self._data + other._data)

    def join(self, other: NluDataset) -> NluDataset:
        """Join this NluDataset with another one."""
        return self + other

    def shuffle(self) -> NluDataset:
        """Shuffles the dataset and returns itself."""
        random.shuffle(self._data)
        return self

    def sample(
        self,
        size: Union[float, int],
        random_state: Optional[int] = None,
        stratify: Any = "intents",
    ) -> NluDataset:
        """
        Sample the dataset preserving intent proportions.

        :param size: can be an integer representing the final number of
            samples or a float representing the ratio
        :param random_state: seed for random processes
        :param stratify: If not None, data is split in a stratified fashion, using
            this as the class labels. Default is using the intent labels for
            stratification.
        """
        if isinstance(size, numbers.Integral):
            size = size / self.n_samples

        _, sampled = self.train_test_split(
            test_size=size, random_state=random_state, stratify=stratify
        )
        return sampled

    def filter_by_intent_name(
        self,
        excluded: Optional[List[str]] = None,
        allowed: Optional[List[str]] = None,
    ) -> NluDataset:
        """
        Filter the dataset by intents.

        :param excluded: List of intents for which not to return data
        :param allowed: List of intents for which to return data
        """
        if excluded is None:
            excluded = []
        if allowed is None:
            allowed = self.unique_intents

        data = []
        for intent, record in zip(self.intents, self._data):
            if intent in excluded:
                continue
            if intent not in allowed:
                continue
            data.append(record)
        return NluDataset._from_data(data)

    def select_first_n_intents(self, n: int) -> NluDataset:
        """
        Return a dataset with a max of `n` intents from the dataset.

        The order of the selected intents is given by
        `self.unique_intents`

        :param n: Number of intents to return
        """
        allowed = self.unique_intents[:n]
        ds = self.filter_by_intent_name(allowed=allowed)
        return ds

    def clip_by_intent_frequency(
        self, max_frequency: int, min_frequency: Optional[int] = None
    ) -> NluDataset:
        """
        Sample the dataset leaving only `max_freq` samples per intent.

        If min_frequency is given then intents with less than that
        number of samples will be dropped.

        Example:
        -------
        To get a dataset with all intents having the same number of
        samples you can do:

        >>> ds = MyNluDataset()
        >>> ds.intent_frequencies
        Counter({'telefonie_disambiguierung': 562,
                'telefonie_beidseitig': 869,
                'internet_kein_internet': 815,
                'internet_langsam': 764,
                'internet_abbrueche': 654,
                'komplettausfall_eindeutig': 931})

        >>> ds.clip_by_intent_frequency(700, 700).intent_frequencies
        Counter({'telefonie_beidseitig': 700,
                'internet_kein_internet': 700,
                'internet_langsam': 700,
                'komplettausfall_eindeutig': 700})


        :param max_frequency: maximum number of samples per intent to
            include in the returned dataset
        :param min_frequency: minimum number of samples per intent to
            include in the returned dataset. If an intent has less
            samples, it will be dropped.

        """
        freqs: collections.Counter = collections.Counter()
        data = []
        for intent, record in zip(self.intents, self._data):
            if (
                min_frequency is not None
                and self.intent_frequencies[intent] <= min_frequency
            ):
                continue

            if freqs[intent] == max_frequency:
                continue

            freqs[intent] += 1
            data.append(record)
        return NluDataset._from_data(data)

    def subsample_by_intent_frequency(
        self,
        target_rate: float,
        min_frequency: int,
    ) -> NluDataset:
        """
        Return a smaller dataset with similar intent distribution.

        Only a certain percentage of each intent's data is kept, while
        for intents with fewer examples that percentage is higher.

        Intents with less than min_frequency samples will not be
        reduced. For all others, the target count is computed as
        `min_freq + (num_samples - min_freq) * target_rate`

        :param target_rate: fraction of data to keep beyond min_frequency
        :param min_frequency: number of utterance that are always kept
        """
        freqs: collections.Counter = collections.Counter()
        data = []
        for intent, record in zip(self.intents, self._data):
            intent_freq = self.intent_frequencies[intent]
            target_freq = min_frequency + ((intent_freq - min_frequency) * target_rate)
            is_below_min = intent_freq <= min_frequency
            is_below_target = freqs[intent] < target_freq
            if is_below_min or is_below_target:
                freqs[intent] += 1
                data.append(record)
        return NluDataset._from_data(data)

    def train_test_split(
        self, stratify="intents", **args
    ) -> Tuple[NluDataset, NluDataset]:
        """
        Split dataset into train and test partitions.

        Uses sklearn.model_selection.train_test_split under the hood. Arguments are
        directly passed to the sklearn function, with the exception of `stratify`: this
        is set by default to "intents" which will set stratify to self.intents (i.e.,
        stratify by intents). Otherwise, stratify works like with sklearn.

        :param stratify: like sklearn train_test_split argument `stratify`.
            Defaults to using self.intents.
        :param args: additional arguments for sklearn's train_test_split, see
            https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html
        """
        if stratify == "intents":
            stratify = self.intents
        (data_train, data_test) = train_test_split(
            self._data,
            stratify=stratify,
            **args,
        )
        train_ds = NluDataset._from_data(data_train)
        test_ds = NluDataset._from_data(data_test)
        return train_ds, test_ds

    def cross_validation_splits(
        self, cv_iterator: BaseCrossValidator = None
    ) -> Iterator[Tuple[NluDataset, NluDataset]]:
        """
        Cross validation generator function.

        :param cv_iterator: sklearn cross validator object to control cv folds
        """
        if cv_iterator is None:
            cv_iterator = StratifiedKFold(n_splits=5, shuffle=True)
        for train_idx, test_idx in cv_iterator.split(self.texts, self.intents):
            train_ds = self[train_idx]
            test_ds = self[test_idx]
            yield train_ds, test_ds

    def to_records(self):
        """Deprectaed. Use to_json() instead."""
        warnings.warn(
            "This method will be removed in a future version. Please use "
            "NluDataset.to_json().",
            DeprecationWarning,
        )
        return self.to_json()

    def to_dict(self) -> Dict[str, list]:
        """
        Return intent data as dict.

        Intents are keys, each holding a list of corresponding
        utterances.
        """
        d = collections.defaultdict(list)
        for record in self._data:
            d[record[1]].append(record[0])
        return dict(d)

    def to_json(self, path: Optional[Union[Path, str]] = None) -> Optional[List[Dict]]:
        """
        Convert dataset to JSON.

        If a path is given, saves to that file. Otherwise returns records as a list of
        dicts.

        :param path: optional path under which to save the JSON file
        """
        records = [
            {"text": text, "intent": intent, "entities": entities, "n_best_list": nbest}
            for text, intent, entities, nbest in self._data
        ]

        if path is None:
            return records

        with open(path, "w") as f:
            json.dump(records, f, indent=4, ensure_ascii=False)
        return None


def concat(dataset_list: List[NluDataset]) -> NluDataset:
    """
    Concatenate NluDatasets.

    :param dataset_list: list of NluDataset objects
    """
    joined = NluDataset([])
    for ds in dataset_list:
        joined += ds
    return joined



# Copyright (c) 2021 Yaser Martinez-Palenzuela, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

import logging
import os
import re
import tempfile

from fasttext import train_supervised

from .vendor import Vendor


logger = logging.getLogger(__name__)
DEFAULT_CONFIG = {
    "wordNgrams": 1,
    "dim": 100,
    "verbose": 2,
    "ws": 5,
    "neg": 10,
    "bucket": 50000,
    "loss": "softmax",
    "minn": 4,
    "maxn": 5,
    "t": 0.01,
    "minCount": 1,
}


class FastText(Vendor):
    def __init__(self, epochs=10000, lr=0.1, config=DEFAULT_CONFIG):
        """
        Interface for the FastText classifier.

        See `FastText classifier <https://github.com/facebookresearch/fastText/>`_.

        Features:
            * bag of n-grams as additional features to capture some
              partial information about the local word order
            * fast and memory efficient mapping of the n-grams by using
              the hashing trick
            * Leverages subword information (via character n-gram)
            * Can build representations for out of vocabulary tokens
              (char ngrams average)
            * Fast training with multicore support implementation

        Pros:
            * Competitive performance, specially on low data regime
            * Efficient both at training and at test time
            * Possibility of incorporating domain knowledge
            * Open source

        **Reference:**

        `Paper: <https://arxiv.org/abs/1607.01759>`_

        :param epochs: training epochs
        :type epochs: int
        :param lr: learning rate
        :type lr: float
        :param config: dictionary with additional parameters
        :type config: dict
        """
        self._alias = self.name
        self._epochs = epochs
        self._lr = lr
        self._model = None
        self._config = config

    def train_intent(self, dataset):
        """Train intent classifier."""
        train_data = self._convert(dataset)
        logger.info(f"Training on {dataset.n_samples} samples")
        self._model = train_supervised(
            input=train_data, epoch=self._epochs, lr=self._lr, **self._config
        )
        # remove tempfile
        os.remove(train_data)
        return self

    def test_intent(self, dataset, return_probs=False):
        """Test intent classifier."""
        logger.info(f"Testing on {dataset.n_samples} samples")
        intents = []
        probs = []
        for text in dataset.texts:
            text = self._clean_text(text)
            text = "".join(text.splitlines())
            result = self._model.predict(text)
            intent = result[0][0]
            intent = intent[len("__label__") :]
            prob = result[1][0]
            intents.append(intent)
            probs.append(prob)
        if return_probs:
            return intents, probs
        return intents

    def _convert(self, dataset):
        lines = []
        for text, intent in zip(dataset.texts, dataset.intents):
            text = self._clean_text(text)
            line = f"__label__{intent} {text}"
            lines.append(line)
        lines = "\n".join(lines)

        with tempfile.NamedTemporaryFile(mode="w", encoding="utf8", delete=False) as f:
            f.write(lines)
            train_data = f.name

        return train_data

    @staticmethod
    def _clean_text(text):
        return re.sub("[^a-zA-Z\n.]", " ", text)



from packaging import version
from test_datasets import entities as sample_entities
from test_datasets import intents as sample_intents
from test_datasets import texts as sample_texts

from nlubridge import EntityKeys, NBestKeys, NluDataset


# Following functions are run by all vendor tests to test
# functionalities shared by all
# -----------------------------


def assert_preds_are_intents(vendor, unique_intents):
    test_ds = NluDataset(["Ich habe kein DSL und telefon"])
    preds = vendor.test_intent(test_ds)
    assert isinstance(preds, list)
    assert len(preds) == 1
    assert preds[0] in unique_intents


def assert_return_probs(vendor, unique_intents):
    test_ds = NluDataset(["Ich habe kein DSL und telefon"])
    preds, probs = vendor.test_intent(test_ds, return_probs=True)
    assert preds[0] in unique_intents
    assert isinstance(probs[0], float) or (probs[0] == 1)


def assert_multiple_utterances_predicted(vendor, train_data):
    preds, probs = vendor.test_intent(train_data[:10], return_probs=True)
    assert len(preds) == 10
    assert isinstance(preds[0], str)
    assert len(probs) == 10
    assert isinstance(probs[0], float) or (probs[0] == 1)


# Following are tests for each vendor. We bundled test by vendor so we
# can test an individual vendor independently during development
# -----------------------------


def test_vendor_attributes():
    from nlubridge.vendors.tfidf_intent_classifier import TfidfIntentClassifier

    clf = TfidfIntentClassifier()
    assert clf.name == "TfidfIntentClassifier"
    assert clf.alias == "TfidfIntentClassifier"
    clf.alias = "tfidf"
    assert clf.alias == "tfidf"


def test_tfidf(train_data):
    from nlubridge.vendors.tfidf_intent_classifier import TfidfIntentClassifier

    # test initialization
    bow = TfidfIntentClassifier()

    # test train_intent()
    bow.train_intent(train_data)

    # test test_intent()
    assert_preds_are_intents(bow, train_data.unique_intents)
    assert_return_probs(bow, train_data.unique_intents)
    assert_multiple_utterances_predicted(bow, train_data)


def test_rasa(train_data):
    from rasa import __version__ as rasa_version

    if version.parse(rasa_version) < version.parse("3.0.0"):
        from nlubridge.vendors.rasa2 import Rasa2

        rasa = Rasa2()
    else:
        from nlubridge.vendors.rasa3 import Rasa3

        rasa = Rasa3()

    # test intent classification
    rasa.train_intent(train_data)
    assert_preds_are_intents(rasa, train_data.unique_intents)
    assert_return_probs(rasa, train_data.unique_intents)
    assert_multiple_utterances_predicted(rasa, train_data)

    # test intent + entity classification
    train_ds = NluDataset(sample_texts, sample_intents, sample_entities)
    rasa.train(train_ds)
    test_ds = NluDataset(
        ["I want a flight from Frankfurt to Berlin", "How is the weather in Bonn?"]
    )
    preds = rasa.test(test_ds)
    assert isinstance(preds, NluDataset)
    assert len(preds) == 2
    assert len(preds.confidences) == 2
    assert isinstance(preds.confidences[0], float)
    assert len(preds.n_best_intents) == 2
    assert isinstance(preds.n_best_intents[0], list)
    assert isinstance(preds.n_best_intents[0][0], dict)
    assert preds.n_best_intents[0][0][NBestKeys.INTENT] in train_ds.unique_intents
    assert len(preds.entities) == 2
    assert isinstance(preds.entities[0], list)
    assert isinstance(preds.entities[0][0], dict)
    assert preds.entities[0][0][EntityKeys.TYPE] in train_ds.unique_entities


def test_telekom(train_data):
    from nlubridge.vendors.char_ngram_intent_classifier import CharNgramIntentClassifier

    model = CharNgramIntentClassifier()
    model.train_intent(train_data)
    assert_preds_are_intents(model, train_data.unique_intents)
    assert_return_probs(model, train_data.unique_intents)
    assert_multiple_utterances_predicted(model, train_data)
    # TODO: fix vendor for test: assert_oos_prediction(mediaan)


def test_spacy(train_data):
    from nlubridge.vendors.spacy import Spacy

    # We train with small number of train iterations to speed up tests
    # (performance not important here)
    model = Spacy(n_iter=10)
    model.train_intent(train_data)
    predicted = model.test_intent(train_data)
    assert len(predicted) == len(train_data)
    # assert_preds_are_intents(spacy, train_data.unique_intents)
    # assert_return_probs(spacy, train_data.unique_intents)
    # assert_multiple_utterances_predicted(spacy, train_data)
    # assert_oos_prediction(spacy)


def test_fasttext(train_data):
    from nlubridge.vendors.fasttext import FastText

    fasttext = FastText(epochs=10)
    fasttext.train_intent(train_data)
    assert_preds_are_intents(fasttext, train_data.unique_intents)
    assert_return_probs(fasttext, train_data.unique_intents)
    assert_multiple_utterances_predicted(fasttext, train_data)
